import asyncio
from contextlib import contextmanager
from http import HTTPStatus
import json
from typing import Generator, Optional

from aiohttp import ClientConnectionError
from anyscale_client.exceptions import ApiException
from click import ClickException


class ExpectedLogRetrievalError(Exception):
    pass


class UnsupportedLogRetrievalMethodError(ExpectedLogRetrievalError):
    pass


class LogRetrievalTimeoutError(ExpectedLogRetrievalError):
    pass


@contextmanager
def wrap_as_unsupported_log_retrieval_method_error() -> Generator[None, None, None]:
    """
    Transforms Anyscale's `ApiException` into an expected error type.
    Usually the SDK/CLI client can try the next retrieval method if raised.
    """
    try:
        yield
    except ApiException as e:
        raise UnsupportedLogRetrievalMethodError(
            _get_exception_message(e)
        ) if e.status == HTTPStatus.BAD_REQUEST else e


def _get_exception_message(e: ApiException) -> Optional[str]:
    "Retrieves the `detail` method from any Anyscale `ApiException`"
    return json.loads(e.body).get("error", {}).get("detail")


@contextmanager
def wrap_job_run_log_not_retrievable_on_active_cluster_error(
    job_run_id: str, raise_connection_issue_as_cli_error: bool = False
) -> Generator[None, None, None]:
    """Parses a readable message for connection errors. Transforms connection errors for CLI."""
    try:
        yield
    except (ClientConnectionError, asyncio.TimeoutError) as e:
        error_msg = (
            f'Retrieving logs while the job is running is only supported if the Anyscale client (SDK / CLI) can reach the cluster running the job "{job_run_id}".\n'
            "Check that you are calling this from the same VPN as the cluster, or if the cluster is unreachable because of high load."
        )
        if raise_connection_issue_as_cli_error:
            raise ClickException(error_msg)
        else:
            raise type(e)(error_msg)



import copy
from datetime import datetime, timezone
import json
from pathlib import Path
import re
from threading import Thread
from typing import Any, Dict, Iterator, List, Optional, Tuple
from unittest.mock import Mock, patch
from wsgiref.simple_server import make_server

from google.auth.credentials import AnonymousCredentials
import pytest

from anyscale.client.openapi_client import (
    Cloud,
    ExecuteCommandResponse,
    Organization,
    Project,
    Session,
    SessionCommand,
    SessionCommandTypes,
    SessionListResponse,
    SessionStartingUpData,
    SessionStateData,
    UserInfo,
)
from anyscale.sdk.anyscale_client.models.cluster_compute import ClusterCompute
from anyscale.sdk.anyscale_client.models.cluster_compute_config import (
    ClusterComputeConfig,
)
from anyscale.sdk.anyscale_client.models.compute_template import ComputeTemplate
from anyscale.utils.gcp_utils import GoogleCloudClientFactory


@pytest.fixture()
def base_mock_api_client() -> Mock:
    mock_api_client = Mock()
    return mock_api_client


@pytest.fixture()
def base_mock_anyscale_api_client() -> Mock:
    mock_anyscale_api_client = Mock()
    return mock_anyscale_api_client


@pytest.fixture()
def mock_auth_api_client(
    base_mock_api_client: Mock, base_mock_anyscale_api_client: Mock
):
    mock_auth_api_client = Mock(
        api_client=base_mock_api_client,
        anyscale_api_client=base_mock_anyscale_api_client,
        host="https://api.anyscale.com",
    )
    with patch.multiple(
        "anyscale.controllers.base_controller",
        get_auth_api_client=Mock(return_value=mock_auth_api_client),
    ):
        yield


@pytest.fixture()
def mock_api_client_with_session(
    base_mock_api_client: Mock, session_test_data: Session
) -> Mock:
    base_mock_api_client.list_sessions_api_v2_sessions_get.return_value = SessionListResponse(
        results=[session_test_data]
    )
    return base_mock_api_client


@pytest.fixture(scope="module")
def cloud_test_data() -> Cloud:
    return Cloud(
        id="cloud_id_1",
        name="cloud_name_1",
        provider="AWS",
        region="region",
        credentials="credentials",
        creator_id="creator_id",
        type="PUBLIC",
        created_at=datetime.now(timezone.utc),
        config='{"max_stopped_instances": 0}',
        state="ACTIVE",
    )


@pytest.fixture(scope="module")
def project_test_data() -> Project:
    return Project(
        name="project_name",
        description="test project",
        cloud_id="cloud_id",
        initial_cluster_config="initial_config",
        id="project_id",
        created_at=datetime.now(tz=timezone.utc),
        creator_id="creator_id",
        is_owner=True,
        directory_name="/directory/name",
        active_sessions=0,
        last_activity_at=datetime.now(tz=timezone.utc),
        is_default=False,
    )


@pytest.fixture(scope="module")
def session_test_data() -> Session:
    return Session(
        id="session_id",
        name="session_name",
        created_at=datetime.now(tz=timezone.utc),
        snapshots_history=[],
        tensorboard_available=False,
        project_id="project_id",
        state="Running",
        idle_timeout=120,
        access_token="token",
        host_name="https://test.anyscale.com",
    )


@pytest.fixture(scope="module")
def compute_config_test_data() -> ClusterComputeConfig:
    return ClusterComputeConfig(
        cloud_id="cld_123",
        head_node_type={
            "instance_type": "m5.2xlarge",
            "name": "head-node-type",
            "resources": None,
        },
        max_workers=20,
        region="us-west-2",
        worker_node_types=[
            {
                "instance_type": "m5.4xlarge",
                "max_workers": 10,
                "min_workers": None,
                "name": "worker-node-type-0",
                "resources": None,
                "use_spot": False,
            },
            {
                "instance_type": "g4dn.4xlarge",
                "max_workers": 10,
                "min_workers": None,
                "name": "worker-node-type-1",
                "resources": None,
                "use_spot": False,
            },
        ],
    )


@pytest.fixture(scope="module")
def cluster_compute_test_data(
    compute_config_test_data: ClusterComputeConfig,
) -> ClusterCompute:
    return ClusterCompute(
        id="cpt_123",
        name="compute_config_name",
        creator_id="user_123",
        organization_id="org_123",
        project_id="prj_123",
        created_at=datetime.now(tz=timezone.utc),
        last_modified_at=datetime.now(tz=timezone.utc),
        config=compute_config_test_data,
        anonymous=False,
        version=1,
    )


@pytest.fixture(scope="module")
def compute_template_test_data(
    compute_config_test_data: ClusterComputeConfig,
) -> ComputeTemplate:
    return ComputeTemplate(
        id="cpt_123",
        name="compute_config_name",
        creator_id="user_123",
        organization_id="org_123",
        project_id="prj_123",
        created_at=datetime.now(tz=timezone.utc),
        last_modified_at=datetime.now(tz=timezone.utc),
        config=compute_config_test_data,
        anonymous=False,
        version=1,
    )


@pytest.fixture(scope="module")
def session_start_error_test_data() -> Session:
    return Session(
        id="session_id",
        name="session_name",
        created_at=datetime.now(tz=timezone.utc),
        snapshots_history=[],
        tensorboard_available=False,
        project_id="project_id",
        state="StartupErrored",
        state_data=SessionStateData(
            startup=SessionStartingUpData(startup_error="start up error")
        ),
        idle_timeout=120,
        access_token="token",
    )


@pytest.fixture(scope="module")
def session_terminated_test_data() -> Session:
    return Session(
        id="session_id",
        name="session_name",
        created_at=datetime.now(tz=timezone.utc),
        snapshots_history=[],
        tensorboard_available=False,
        project_id="project_id",
        state="Terminated",
        idle_timeout=120,
        access_token="token",
    )


@pytest.fixture(scope="module")
def session_command_test_data() -> SessionCommand:
    return SessionCommand(
        id="session_command_id",
        created_at=datetime.now(tz=timezone.utc),
        name="session_command",
        params="params",
        shell="shell",
        shell_command="shell_command",
        type=SessionCommandTypes.COMMAND_LINE_RUNNER,
    )


@pytest.fixture(scope="module")
def command_id_test_data() -> ExecuteCommandResponse:
    return ExecuteCommandResponse(
        command_id="command_id",
        directory_name="dir_name",
        dns_address="session.anyscaleuserdata-dev.com",
    )


@pytest.fixture(scope="module")
def userinfo_test_data() -> UserInfo:
    return UserInfo(
        id="mock_user_id",
        email="mock_email",
        name="mock_name",
        username="mock_user_name",
        verified=True,
        organization_permission_level="owner",
        organization_ids=["mock_org_id_1"],
        organizations=[
            Organization(
                id="mock_org_id_1",
                name="mock_org_1",
                public_identifier="mock_org_1",
                default_cloud_id="mock_default_cloud_id",
                sso_required=False,
                is_general_platform_enabled=True,
                is_private_endpoints_enabled=False,
            )
        ],
        ld_hash="mock_ld_hash",
        ld_hash_fields=["email", "name", "username", "organization_ids"],
    )


@pytest.fixture(scope="session", autouse=True)
def default_session_fixture(
    request: "_pytest.fixtures.SubRequest",  # type: ignore # noqa: F821
) -> Any:
    with patch("time.sleep", Mock()):
        yield


class RequestTracker:
    def __init__(self):
        self.responses: Dict[str, List[Tuple[int, Optional[str]]]] = {}
        self.seen_requests: List[Any] = []

    def reset(self, responses):
        for k, v in responses.items():
            re.compile(k)
            assert all(len(i) == 2 for i in v)
        self.responses = copy.deepcopy(responses)
        self.seen_requests = []


class GCloudMockHandler:
    def __init__(self, tracker, *args, **kwargs):
        self.tracker = tracker

    def __call__(self, env, start_response):
        self.tracker.seen_requests.append(env)

        path = env["PATH_INFO"]

        for regex in self.tracker.responses:
            if re.match(regex, path):
                code, body_file = self.tracker.responses[regex].pop(0)
                start_response(code, [("Content-Type", "application/json")])
                if body_file is not None:
                    with open(
                        Path(__file__).parent.joinpath("gcp_responses", body_file)
                    ) as f:
                        return [json.dumps(json.load(f)).strip().encode()]
                return []

        pytest.fail(
            f"Un handled request to {self.path}:\n{self.tracker.responses}",
            pytrace=False,
        )


@pytest.fixture()
def setup_mock_server() -> Iterator[Tuple[GoogleCloudClientFactory, RequestTracker]]:
    tracker = RequestTracker()
    server = make_server("localhost", 0, GCloudMockHandler(tracker))
    port = server.server_port
    print(f"Serving on (http://localhost:{port})")
    t = Thread(target=server.serve_forever, daemon=True)
    t.start()
    yield (
        GoogleCloudClientFactory(
            credentials=AnonymousCredentials(),
            force_rest=True,
            client_options={"api_endpoint": f"http://127.0.0.1:{port}"},
        ),
        tracker,
    )
    print(f"Shutting down server on (http://localhost:{port})")
    server.server_close()
    t.join(timeout=0.1)



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class PlatformFineTuningJob(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'result_files': 'list[str]',
        'trained_tokens': 'int',
        'hyperparameters': 'CreateFineTuningHyperparameters',
        'training_file': 'str',
        'validation_file': 'str',
        'model': 'str',
        'id': 'str',
        'created_at': 'datetime',
        'finished_at': 'datetime',
        'fine_tuned_model': 'str',
        'status': 'FineTuningJobStatus',
        'error': 'str',
        'creator_id': 'str',
        'cloud_id': 'str',
        'creator_email': 'str',
        'ha_job_id': 'str'
    }

    attribute_map = {
        'result_files': 'result_files',
        'trained_tokens': 'trained_tokens',
        'hyperparameters': 'hyperparameters',
        'training_file': 'training_file',
        'validation_file': 'validation_file',
        'model': 'model',
        'id': 'id',
        'created_at': 'created_at',
        'finished_at': 'finished_at',
        'fine_tuned_model': 'fine_tuned_model',
        'status': 'status',
        'error': 'error',
        'creator_id': 'creator_id',
        'cloud_id': 'cloud_id',
        'creator_email': 'creator_email',
        'ha_job_id': 'ha_job_id'
    }

    def __init__(self, result_files=[], trained_tokens=None, hyperparameters=None, training_file=None, validation_file=None, model=None, id=None, created_at=None, finished_at=None, fine_tuned_model=None, status=None, error=None, creator_id=None, cloud_id=None, creator_email=None, ha_job_id=None, local_vars_configuration=None):  # noqa: E501
        """PlatformFineTuningJob - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._result_files = None
        self._trained_tokens = None
        self._hyperparameters = None
        self._training_file = None
        self._validation_file = None
        self._model = None
        self._id = None
        self._created_at = None
        self._finished_at = None
        self._fine_tuned_model = None
        self._status = None
        self._error = None
        self._creator_id = None
        self._cloud_id = None
        self._creator_email = None
        self._ha_job_id = None
        self.discriminator = None

        if result_files is not None:
            self.result_files = result_files
        if trained_tokens is not None:
            self.trained_tokens = trained_tokens
        if hyperparameters is not None:
            self.hyperparameters = hyperparameters
        self.training_file = training_file
        if validation_file is not None:
            self.validation_file = validation_file
        self.model = model
        self.id = id
        self.created_at = created_at
        if finished_at is not None:
            self.finished_at = finished_at
        self.fine_tuned_model = fine_tuned_model
        self.status = status
        if error is not None:
            self.error = error
        self.creator_id = creator_id
        self.cloud_id = cloud_id
        self.creator_email = creator_email
        if ha_job_id is not None:
            self.ha_job_id = ha_job_id

    @property
    def result_files(self):
        """Gets the result_files of this PlatformFineTuningJob.  # noqa: E501

        Array with the ID of the .json file that contains finetuning results. Available after the job is `running`  # noqa: E501

        :return: The result_files of this PlatformFineTuningJob.  # noqa: E501
        :rtype: list[str]
        """
        return self._result_files

    @result_files.setter
    def result_files(self, result_files):
        """Sets the result_files of this PlatformFineTuningJob.

        Array with the ID of the .json file that contains finetuning results. Available after the job is `running`  # noqa: E501

        :param result_files: The result_files of this PlatformFineTuningJob.  # noqa: E501
        :type: list[str]
        """

        self._result_files = result_files

    @property
    def trained_tokens(self):
        """Gets the trained_tokens of this PlatformFineTuningJob.  # noqa: E501

        Number of tokens trained on. Available after the job is `succeeded`  # noqa: E501

        :return: The trained_tokens of this PlatformFineTuningJob.  # noqa: E501
        :rtype: int
        """
        return self._trained_tokens

    @trained_tokens.setter
    def trained_tokens(self, trained_tokens):
        """Sets the trained_tokens of this PlatformFineTuningJob.

        Number of tokens trained on. Available after the job is `succeeded`  # noqa: E501

        :param trained_tokens: The trained_tokens of this PlatformFineTuningJob.  # noqa: E501
        :type: int
        """

        self._trained_tokens = trained_tokens

    @property
    def hyperparameters(self):
        """Gets the hyperparameters of this PlatformFineTuningJob.  # noqa: E501

        Hyperparameters used for fine-tuning.  # noqa: E501

        :return: The hyperparameters of this PlatformFineTuningJob.  # noqa: E501
        :rtype: CreateFineTuningHyperparameters
        """
        return self._hyperparameters

    @hyperparameters.setter
    def hyperparameters(self, hyperparameters):
        """Sets the hyperparameters of this PlatformFineTuningJob.

        Hyperparameters used for fine-tuning.  # noqa: E501

        :param hyperparameters: The hyperparameters of this PlatformFineTuningJob.  # noqa: E501
        :type: CreateFineTuningHyperparameters
        """

        self._hyperparameters = hyperparameters

    @property
    def training_file(self):
        """Gets the training_file of this PlatformFineTuningJob.  # noqa: E501

        ID of the .jsonl file used for training  # noqa: E501

        :return: The training_file of this PlatformFineTuningJob.  # noqa: E501
        :rtype: str
        """
        return self._training_file

    @training_file.setter
    def training_file(self, training_file):
        """Sets the training_file of this PlatformFineTuningJob.

        ID of the .jsonl file used for training  # noqa: E501

        :param training_file: The training_file of this PlatformFineTuningJob.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and training_file is None:  # noqa: E501
            raise ValueError("Invalid value for `training_file`, must not be `None`")  # noqa: E501

        self._training_file = training_file

    @property
    def validation_file(self):
        """Gets the validation_file of this PlatformFineTuningJob.  # noqa: E501

        ID of the .jsonl file used for validation. Optional  # noqa: E501

        :return: The validation_file of this PlatformFineTuningJob.  # noqa: E501
        :rtype: str
        """
        return self._validation_file

    @validation_file.setter
    def validation_file(self, validation_file):
        """Sets the validation_file of this PlatformFineTuningJob.

        ID of the .jsonl file used for validation. Optional  # noqa: E501

        :param validation_file: The validation_file of this PlatformFineTuningJob.  # noqa: E501
        :type: str
        """

        self._validation_file = validation_file

    @property
    def model(self):
        """Gets the model of this PlatformFineTuningJob.  # noqa: E501

        A supported model used as a base for fine-tuning  # noqa: E501

        :return: The model of this PlatformFineTuningJob.  # noqa: E501
        :rtype: str
        """
        return self._model

    @model.setter
    def model(self, model):
        """Sets the model of this PlatformFineTuningJob.

        A supported model used as a base for fine-tuning  # noqa: E501

        :param model: The model of this PlatformFineTuningJob.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and model is None:  # noqa: E501
            raise ValueError("Invalid value for `model`, must not be `None`")  # noqa: E501

        self._model = model

    @property
    def id(self):
        """Gets the id of this PlatformFineTuningJob.  # noqa: E501

        The ID of the fine-tuning job  # noqa: E501

        :return: The id of this PlatformFineTuningJob.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this PlatformFineTuningJob.

        The ID of the fine-tuning job  # noqa: E501

        :param id: The id of this PlatformFineTuningJob.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def created_at(self):
        """Gets the created_at of this PlatformFineTuningJob.  # noqa: E501

        The time the job was submitted  # noqa: E501

        :return: The created_at of this PlatformFineTuningJob.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this PlatformFineTuningJob.

        The time the job was submitted  # noqa: E501

        :param created_at: The created_at of this PlatformFineTuningJob.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def finished_at(self):
        """Gets the finished_at of this PlatformFineTuningJob.  # noqa: E501

        The time the job finished  # noqa: E501

        :return: The finished_at of this PlatformFineTuningJob.  # noqa: E501
        :rtype: datetime
        """
        return self._finished_at

    @finished_at.setter
    def finished_at(self, finished_at):
        """Sets the finished_at of this PlatformFineTuningJob.

        The time the job finished  # noqa: E501

        :param finished_at: The finished_at of this PlatformFineTuningJob.  # noqa: E501
        :type: datetime
        """

        self._finished_at = finished_at

    @property
    def fine_tuned_model(self):
        """Gets the fine_tuned_model of this PlatformFineTuningJob.  # noqa: E501

        A generated ID of the fine-tuned model  # noqa: E501

        :return: The fine_tuned_model of this PlatformFineTuningJob.  # noqa: E501
        :rtype: str
        """
        return self._fine_tuned_model

    @fine_tuned_model.setter
    def fine_tuned_model(self, fine_tuned_model):
        """Sets the fine_tuned_model of this PlatformFineTuningJob.

        A generated ID of the fine-tuned model  # noqa: E501

        :param fine_tuned_model: The fine_tuned_model of this PlatformFineTuningJob.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and fine_tuned_model is None:  # noqa: E501
            raise ValueError("Invalid value for `fine_tuned_model`, must not be `None`")  # noqa: E501

        self._fine_tuned_model = fine_tuned_model

    @property
    def status(self):
        """Gets the status of this PlatformFineTuningJob.  # noqa: E501

        The current status of the job  # noqa: E501

        :return: The status of this PlatformFineTuningJob.  # noqa: E501
        :rtype: FineTuningJobStatus
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this PlatformFineTuningJob.

        The current status of the job  # noqa: E501

        :param status: The status of this PlatformFineTuningJob.  # noqa: E501
        :type: FineTuningJobStatus
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    @property
    def error(self):
        """Gets the error of this PlatformFineTuningJob.  # noqa: E501

        Error message. Available if the job is `failed`  # noqa: E501

        :return: The error of this PlatformFineTuningJob.  # noqa: E501
        :rtype: str
        """
        return self._error

    @error.setter
    def error(self, error):
        """Sets the error of this PlatformFineTuningJob.

        Error message. Available if the job is `failed`  # noqa: E501

        :param error: The error of this PlatformFineTuningJob.  # noqa: E501
        :type: str
        """

        self._error = error

    @property
    def creator_id(self):
        """Gets the creator_id of this PlatformFineTuningJob.  # noqa: E501

        The ID of the user who submitted the job  # noqa: E501

        :return: The creator_id of this PlatformFineTuningJob.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this PlatformFineTuningJob.

        The ID of the user who submitted the job  # noqa: E501

        :param creator_id: The creator_id of this PlatformFineTuningJob.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_id is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_id`, must not be `None`")  # noqa: E501

        self._creator_id = creator_id

    @property
    def cloud_id(self):
        """Gets the cloud_id of this PlatformFineTuningJob.  # noqa: E501

        Cloud id fo the FT job  # noqa: E501

        :return: The cloud_id of this PlatformFineTuningJob.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this PlatformFineTuningJob.

        Cloud id fo the FT job  # noqa: E501

        :param cloud_id: The cloud_id of this PlatformFineTuningJob.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_id`, must not be `None`")  # noqa: E501

        self._cloud_id = cloud_id

    @property
    def creator_email(self):
        """Gets the creator_email of this PlatformFineTuningJob.  # noqa: E501

        Email of user who triggered the fine tuning job.  # noqa: E501

        :return: The creator_email of this PlatformFineTuningJob.  # noqa: E501
        :rtype: str
        """
        return self._creator_email

    @creator_email.setter
    def creator_email(self, creator_email):
        """Sets the creator_email of this PlatformFineTuningJob.

        Email of user who triggered the fine tuning job.  # noqa: E501

        :param creator_email: The creator_email of this PlatformFineTuningJob.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_email is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_email`, must not be `None`")  # noqa: E501

        self._creator_email = creator_email

    @property
    def ha_job_id(self):
        """Gets the ha_job_id of this PlatformFineTuningJob.  # noqa: E501


        :return: The ha_job_id of this PlatformFineTuningJob.  # noqa: E501
        :rtype: str
        """
        return self._ha_job_id

    @ha_job_id.setter
    def ha_job_id(self, ha_job_id):
        """Sets the ha_job_id of this PlatformFineTuningJob.


        :param ha_job_id: The ha_job_id of this PlatformFineTuningJob.  # noqa: E501
        :type: str
        """

        self._ha_job_id = ha_job_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, PlatformFineTuningJob):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, PlatformFineTuningJob):
            return True

        return self.to_dict() != other.to_dict()



import datetime
from unittest.mock import Mock, patch

from click.exceptions import ClickException
from freezegun import freeze_time
import pytest

from anyscale.client.openapi_client import Session, SessionListResponse
from anyscale.client.openapi_client.models.project import Project
from anyscale.util import (
    credentials_check_sanity,
    deserialize_datetime,
    get_project_directory_name,
    get_working_dir,
    wait_for_session_start,
)


def test_deserialize_datetime() -> None:
    date_str = "2020-07-02T20:16:04.000000+00:00"
    assert deserialize_datetime(date_str) == datetime.datetime(
        2020, 7, 2, 20, 16, 4, tzinfo=datetime.timezone.utc
    )


def test_wait_for_session_start(
    mock_api_client_with_session: Mock, session_test_data: Session
) -> None:
    result = wait_for_session_start(
        session_test_data.project_id, session_test_data.id, mock_api_client_with_session
    )
    assert result == session_test_data.id


def test_wait_for_session_start_error(
    mock_api_client_with_session: Mock, session_start_error_test_data: Session
) -> None:
    mock_api_client_with_session.list_sessions_api_v2_sessions_get.return_value = SessionListResponse(
        results=[session_start_error_test_data]
    )
    with pytest.raises(ClickException) as e:
        wait_for_session_start(
            session_start_error_test_data.project_id,
            session_start_error_test_data.id,
            mock_api_client_with_session,
        )

    assert "Error while starting cluster" in e.value.message
    mock_api_client_with_session.list_sessions_api_v2_sessions_get.assert_called_once_with(
        project_id=session_start_error_test_data.project_id,
        name=session_start_error_test_data.id,
        active_only=False,
    )


def test_wait_for_session_start_terminated(
    mock_api_client_with_session: Mock, session_terminated_test_data: Session
) -> None:
    mock_api_client_with_session.list_sessions_api_v2_sessions_get.return_value = SessionListResponse(
        results=[session_terminated_test_data]
    )

    # Mock time so we don't need to wait 30 real time seconds for terminated
    # grace period
    freezer = freeze_time("2023-01-1 12:00:00")
    freezer.start()

    def sleep_mock(*args, **kwargs):
        freezer.stop()

    with patch("time.sleep", sleep_mock), pytest.raises(ClickException) as e:
        wait_for_session_start(
            session_terminated_test_data.project_id,
            session_terminated_test_data.id,
            mock_api_client_with_session,
        )

    assert "Cluster is still in stopped/terminated state" in e.value.message
    mock_api_client_with_session.list_sessions_api_v2_sessions_get.assert_called_with(
        project_id=session_terminated_test_data.project_id,
        name=session_terminated_test_data.id,
        active_only=False,
    )


def test_wait_for_session_start_terminated_grace_period(
    mock_api_client_with_session: Mock,
    session_terminated_test_data: Session,
    session_test_data: Session,
) -> None:
    mock_api_client_with_session.list_sessions_api_v2_sessions_get.return_value = SessionListResponse(
        results=[session_terminated_test_data]
    )
    """
    Tests first iteration of loop with session in terminated state. On next iteration, we update
    the session to be in running state. Since the terminated state happens within the grace period,
    there should be no error (as opposed to test_wait_for_session_start_terminated)
    """

    def sleep_mock(*args, **kwargs):
        mock_api_client_with_session.list_sessions_api_v2_sessions_get.assert_called_with(
            project_id=session_terminated_test_data.project_id,
            name=session_terminated_test_data.id,
            active_only=False,
        )
        # Changed mocked response from terminated to running
        mock_api_client_with_session.list_sessions_api_v2_sessions_get.return_value = SessionListResponse(
            results=[session_test_data]
        )
        mock_api_client_with_session.reset_mock()

    with patch("time.sleep", sleep_mock):
        result = wait_for_session_start(
            session_terminated_test_data.project_id,
            session_terminated_test_data.id,
            mock_api_client_with_session,
        )

    assert result == session_test_data.id

    mock_api_client_with_session.list_sessions_api_v2_sessions_get.assert_called_with(
        project_id=session_test_data.project_id,
        name=session_test_data.id,
        active_only=False,
    )


def test_wait_for_session_start_missing_session(
    mock_api_client_with_session: Mock, session_start_error_test_data: Session
) -> None:
    mock_api_client_with_session.list_sessions_api_v2_sessions_get.return_value = SessionListResponse(
        results=[]
    )
    with pytest.raises(ClickException) as e:
        wait_for_session_start(
            session_start_error_test_data.project_id,
            session_start_error_test_data.id,
            mock_api_client_with_session,
        )

    assert "Cluster doesn't exist" in e.value.message
    mock_api_client_with_session.list_sessions_api_v2_sessions_get.assert_called_once_with(
        project_id=session_start_error_test_data.project_id,
        name=session_start_error_test_data.id,
        active_only=False,
    )


def test_get_project_directory_name(project_test_data: Project) -> None:
    mock_api_client = Mock()
    mock_api_client.get_project_api_v2_projects_project_id_get.return_value.result.directory_name = (
        project_test_data.directory_name
    )

    dir_name = get_project_directory_name(project_test_data.id, mock_api_client)

    assert dir_name == project_test_data.directory_name
    mock_api_client.get_project_api_v2_projects_project_id_get.assert_called_once_with(
        project_test_data.id
    )


def test_get_working_dir(project_test_data: Project) -> None:
    mock_api_client = Mock()
    mock_api_client.get_project_api_v2_projects_project_id_get.return_value.result.directory_name = (
        project_test_data.directory_name
    )

    working_dir = get_working_dir({}, project_test_data.id, mock_api_client)
    assert working_dir == f"/home/ray/{project_test_data.directory_name}"

    working_dir = get_working_dir(
        {"metadata": {"anyscale": {"working_dir": "test_working_dir"}}},
        project_test_data.id,
        mock_api_client,
    )
    assert working_dir == "test_working_dir"


def test_credentials_check_sanity():
    # valid credentials
    assert credentials_check_sanity("sss_abcdefg1234567")
    assert credentials_check_sanity("aph0_abcdefg1234567")
    assert credentials_check_sanity("ash0_abcdefg1234567")
    assert credentials_check_sanity("sss_this-=!./#should_3$be_a_valid_ToKEn.%")
    assert credentials_check_sanity("aph0_052943312512")

    # wrong or no prefix
    assert not credentials_check_sanity("wrong_abcdefg1234567")
    assert not credentials_check_sanity("withoutanyprefixes")



import json
from typing import Any, Dict
from unittest.mock import Mock

from anyscale_client.exceptions import ApiException

from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as BaseApi


class MockBaseApi(Mock, BaseApi):
    pass


def get_mock_base_api(function_to_return_vals: Dict[str, Any]) -> MockBaseApi:
    base_api = Mock()
    for function_name, side_effects in function_to_return_vals.items():
        mock_return_val = Mock()
        if type(side_effects) is list or issubclass(type(side_effects), Exception):
            mock_return_val.side_effect = side_effects
        else:
            mock_return_val.return_value = side_effects
        setattr(base_api, function_name, mock_return_val)
    return base_api


def get_mock_api_exception(status: int, error_message: str) -> ApiException:
    mock_api_exception = ApiException(status=status)
    mock_api_exception.body = json.dumps({"error": {"detail": error_message}})
    return mock_api_exception



from datetime import datetime
import re


def gen_valid_name(prefix: str = "autogenerated") -> str:
    """Generate a name for anyscale resources."""
    INVALID_NAME_REGEX_PATTERN = r"[^A-Za-z0-9_-]"

    generated_name = f"{prefix}-{datetime.now().isoformat()}"
    generated_name = re.sub(INVALID_NAME_REGEX_PATTERN, "-", generated_name)
    return generated_name



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
import itertools
from typing import List, Optional, TypeVar


T = TypeVar("T")


def flatten(*lists: List[T]) -> List[T]:
    return list(itertools.chain(*lists))


def as_list(o: Optional[T]) -> List[T]:
    """Converts "container" holding optional value into a list"""
    return [o] if o else []



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class EditableCloudResource(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'aws_s3_id': 'str',
        'aws_efs_id': 'str',
        'aws_efs_mount_target_ip': 'str',
        'memorydb_cluster_config': 'AWSMemoryDBClusterConfig'
    }

    attribute_map = {
        'aws_s3_id': 'aws_s3_id',
        'aws_efs_id': 'aws_efs_id',
        'aws_efs_mount_target_ip': 'aws_efs_mount_target_ip',
        'memorydb_cluster_config': 'memorydb_cluster_config'
    }

    def __init__(self, aws_s3_id=None, aws_efs_id=None, aws_efs_mount_target_ip=None, memorydb_cluster_config=None, local_vars_configuration=None):  # noqa: E501
        """EditableCloudResource - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._aws_s3_id = None
        self._aws_efs_id = None
        self._aws_efs_mount_target_ip = None
        self._memorydb_cluster_config = None
        self.discriminator = None

        if aws_s3_id is not None:
            self.aws_s3_id = aws_s3_id
        if aws_efs_id is not None:
            self.aws_efs_id = aws_efs_id
        if aws_efs_mount_target_ip is not None:
            self.aws_efs_mount_target_ip = aws_efs_mount_target_ip
        if memorydb_cluster_config is not None:
            self.memorydb_cluster_config = memorydb_cluster_config

    @property
    def aws_s3_id(self):
        """Gets the aws_s3_id of this EditableCloudResource.  # noqa: E501

        The AWS S3 id of this cloud.  # noqa: E501

        :return: The aws_s3_id of this EditableCloudResource.  # noqa: E501
        :rtype: str
        """
        return self._aws_s3_id

    @aws_s3_id.setter
    def aws_s3_id(self, aws_s3_id):
        """Sets the aws_s3_id of this EditableCloudResource.

        The AWS S3 id of this cloud.  # noqa: E501

        :param aws_s3_id: The aws_s3_id of this EditableCloudResource.  # noqa: E501
        :type: str
        """

        self._aws_s3_id = aws_s3_id

    @property
    def aws_efs_id(self):
        """Gets the aws_efs_id of this EditableCloudResource.  # noqa: E501

        The AWS EFS id of this cloud.  # noqa: E501

        :return: The aws_efs_id of this EditableCloudResource.  # noqa: E501
        :rtype: str
        """
        return self._aws_efs_id

    @aws_efs_id.setter
    def aws_efs_id(self, aws_efs_id):
        """Sets the aws_efs_id of this EditableCloudResource.

        The AWS EFS id of this cloud.  # noqa: E501

        :param aws_efs_id: The aws_efs_id of this EditableCloudResource.  # noqa: E501
        :type: str
        """

        self._aws_efs_id = aws_efs_id

    @property
    def aws_efs_mount_target_ip(self):
        """Gets the aws_efs_mount_target_ip of this EditableCloudResource.  # noqa: E501

        The AWS EFS mount target IP of this cloud.  # noqa: E501

        :return: The aws_efs_mount_target_ip of this EditableCloudResource.  # noqa: E501
        :rtype: str
        """
        return self._aws_efs_mount_target_ip

    @aws_efs_mount_target_ip.setter
    def aws_efs_mount_target_ip(self, aws_efs_mount_target_ip):
        """Sets the aws_efs_mount_target_ip of this EditableCloudResource.

        The AWS EFS mount target IP of this cloud.  # noqa: E501

        :param aws_efs_mount_target_ip: The aws_efs_mount_target_ip of this EditableCloudResource.  # noqa: E501
        :type: str
        """

        self._aws_efs_mount_target_ip = aws_efs_mount_target_ip

    @property
    def memorydb_cluster_config(self):
        """Gets the memorydb_cluster_config of this EditableCloudResource.  # noqa: E501

        The AWS memorydb cluster config of this cloud.  # noqa: E501

        :return: The memorydb_cluster_config of this EditableCloudResource.  # noqa: E501
        :rtype: AWSMemoryDBClusterConfig
        """
        return self._memorydb_cluster_config

    @memorydb_cluster_config.setter
    def memorydb_cluster_config(self, memorydb_cluster_config):
        """Sets the memorydb_cluster_config of this EditableCloudResource.

        The AWS memorydb cluster config of this cloud.  # noqa: E501

        :param memorydb_cluster_config: The memorydb_cluster_config of this EditableCloudResource.  # noqa: E501
        :type: AWSMemoryDBClusterConfig
        """

        self._memorydb_cluster_config = memorydb_cluster_config

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, EditableCloudResource):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, EditableCloudResource):
            return True

        return self.to_dict() != other.to_dict()



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
from typing import Union


def ray_semver_compare(a: str, b: str) -> int:
    """
    A dumb implementation of semver comparison because ray versions
    don't actually follow the semver standard.

    The standard for semver versions is "{MAJOR}.{MINOR}.{PATCH}-{PRE_RELEASE_VERSION}," such as 2.7.0 or 2.7.0-oss for Ray versions.
    However, since we use hyphens in the codebase to split and extract information, our standard for Ray versions is "{MAJOR}.{MINOR}.{PATCH}{PRE_RELEASE_VERSION}," like 2.7.0oss without "-" between the patch and the pre-release version.
    Therefore, when comparing Ray versions, we treat "ray2.7.0" as equivalent to "ray2.7.0oss."

    Example:
        ray_semver_compare("ray2.7.0", "ray2.7.0optimized") == 0
        ray_semver_compare("ray2.0.1rc1", "ray2.0.1rc2") == 0
    """
    a_parts = a.split(".")
    b_parts = b.split(".")
    for i, a_part in enumerate(a_parts):
        if i > len(b_parts):
            # No value in part is always less than any string in a part
            return 1

        try:
            parsed_a_part: Union[str, int] = int(a_part)
            parsed_b_part: Union[str, int] = int(b_parts[i])
        except ValueError:
            parsed_a_part = a_part
            parsed_b_part = b_parts[i]

        if parsed_a_part > parsed_b_part:  # type: ignore
            return 1
        elif parsed_a_part < parsed_b_part:  # type: ignore
            return -1
    if len(b_parts) > len(a_parts):
        # B still has parts, it must be bigger
        return -1

    return 0


SUFFIXES = ("optimized", "oss", "nightly")


def get_byod_ray_version_alias(ray_version: str) -> str:
    """
    Get a modified version alias for the Ray version.

    Args:
        ray_version (str): A string representing the Ray framework version (e.g., "2.7.0").

    Returns:
        str: The modified Ray framework version with the "optimized" suffix, if applicable.
    """

    if ray_version.endswith(SUFFIXES) or "rc" in ray_version:
        return ray_version

    if ray_semver_compare(ray_version, "2.7.0") != -1:  # Ray version >= 2.7.0
        return ray_version.replace(ray_version, f"{ray_version}optimized")

    return ray_version



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
import pytest

from shared_anyscale_utils.utils.ray_semver import (
    get_byod_ray_version_alias,
    ray_semver_compare,
)


@pytest.mark.parametrize(
    ("a", "b", "expected_result"),
    [
        ("ray2.7.0", "ray2.7.0", 0),
        ("ray2.7.0optimized", "ray2.7.0", 0),
        ("ray2.7.0oss", "ray2.7.0", 0),
        ("ray2.0.1rc", "ray2.0.1", 0),
        ("ray3.5.1customsuffix", "ray3.5.1", 0),
        ("ray1.2.3.dev456", "ray1.2.3", 0),
        ("ray2.0.0.final", "ray2.0.0", 0),
        ("ray2.7.0", "ray2.6.0", 1),
        ("ray2.0.1rc", "ray2.0.2", -1),
        ("ray2.7.0optimized", "ray2.7.0oss", 0),
    ],
)
def test_ray_semver_compare(a, b, expected_result):
    assert ray_semver_compare(a, b) == expected_result


@pytest.mark.parametrize(
    ("ray_version", "expected_alias"),
    [
        ("2.7.0rc1", "2.7.0rc1"),
        ("2.6.0", "2.6.0"),
        ("2.7.1", "2.7.1optimized"),
        ("2.7.0", "2.7.0optimized"),
        ("nightly", "nightly"),
    ],
)
def test_get_byod_ray_version_alias(ray_version, expected_alias):
    assert get_byod_ray_version_alias(ray_version) == expected_alias



from copy import deepcopy
import json
import os
from typing import Collection, Optional
from unittest.mock import ANY, call, Mock, mock_open, patch

import click
from pydantic import ValidationError
import pytest

from anyscale.client.openapi_client.models.apply_production_service_v2_model import (
    ApplyProductionServiceV2Model,
)
from anyscale.client.openapi_client.models.decoratedlistserviceapimodel_list_response import (
    DecoratedlistserviceapimodelListResponse,
)
from anyscale.client.openapi_client.models.list_response_metadata import (
    ListResponseMetadata,
)
from anyscale.controllers.service_controller import ServiceController
from anyscale.models.service_model import ServiceConfig
from anyscale.utils.workload_types import Workload


TEST_SERVICE_ID = "test-service-id"
TEST_SERVICE_NAME = "test-service-name"


def test_service_config_model_with_entrypoint():
    config_dict = {
        "name": "test_service",
        "project_id": "test_project_id",
        "build_id": "test_build_id",
        "compute_config_id": "test_compute_config_id",
    }
    mock_validate_successful_build = Mock()
    with patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_config = ServiceConfig.parse_obj(config_dict)
        assert service_config.ray_serve_config is None


def test_service_config_model_with_ray_serve_config():
    runtime_env = {
        "pip": ["requests"],
        "working_dir": "https://github.com/anyscale/docs_examples/archive/refs/heads/main.zip",
    }
    config_dict = {
        "name": "test_service",
        "ray_serve_config": {
            "import_path": "serve_hello:entrypoint",
            "runtime_env": runtime_env,
        },
        "project_id": "test_project_id",
        "build_id": "test_build_id",
        "compute_config_id": "test_compute_config_id",
    }
    mock_validate_successful_build = Mock()
    with patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_config = ServiceConfig.parse_obj(config_dict)
        assert service_config.runtime_env is None
        assert service_config.ray_serve_config["runtime_env"] == runtime_env


def test_service_config_model_with_ray_serve_config_working_dir():
    runtime_env = {"pip": ["requests"], "working_dir": "."}
    config_dict = {
        "name": "test_service",
        "ray_serve_config": {
            "import_path": "serve_hello:entrypoint",
            "runtime_env": runtime_env,
        },
        "project_id": "test_project_id",
        "build_id": "test_build_id",
        "compute_config_id": "test_compute_config_id",
    }
    mock_validate_successful_build = Mock()
    with patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_config = ServiceConfig.parse_obj(config_dict)
        assert service_config.ray_serve_config["runtime_env"] == runtime_env
        assert service_config.ray_gcs_external_storage_config is None


def test_service_config_model_with_ray_gcs_external_storage_config():
    config_dict = {
        "name": "test_service",
        "ray_serve_config": {"runtime_env": {"pip": ["requests"], "working_dir": "."}},
        "project_id": "test_project_id",
        "build_id": "test_build_id",
        "compute_config_id": "test_compute_config_id",
        "ray_gcs_external_storage_config": {"address": "addr:port",},
    }
    mock_validate_successful_build = Mock()
    with patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_config = ServiceConfig.parse_obj(config_dict)
        assert (
            service_config.ray_gcs_external_storage_config
            == config_dict["ray_gcs_external_storage_config"]
        )
        assert service_config.runtime_env is None


@pytest.mark.parametrize(
    ("ray_serve_config", "runtime_env"),
    [[{"runtime_env": {"pip": ["requests"], "working_dir": "."}}, None],],
)
def test_generate_config_from_file(
    mock_auth_api_client, ray_serve_config: Optional[str], runtime_env: Optional[str]
):
    """
    This test checks that the config file for Service v2
    can be properly parsed.
    """
    service_controller = ServiceController()

    name = "test_service_name"
    description = "mock_description"
    build_id = "test_build_id"
    compute_config_id = "test_compute_config_id"
    project_id = "test_project_id"
    rollout_strategy = "ROLLOUT"
    canary_percent = 100
    config_dict = {
        "name": name,
        "description": description,
        "ray_serve_config": ray_serve_config,
        "runtime_env": runtime_env,
        "project_id": project_id,
        "build_id": build_id,
        "compute_config_id": compute_config_id,
        "version": "test_abc",
        "canary_percent": canary_percent,
        "rollout_strategy": rollout_strategy,
    }
    mock_validate_successful_build = Mock()
    with patch(
        "builtins.open", mock_open(read_data=json.dumps(config_dict))
    ), patch.multiple("os.path", exists=Mock(return_value=True)), patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
    ):
        config = service_controller.generate_config_from_file(
            "test_ser8vice_config_file",
        )

        expected_service_config = ServiceConfig(
            name=name,
            description=description,
            ray_serve_config=ray_serve_config,
            runtime_env=runtime_env,
            project_id=project_id,
            build_id=build_id,
            compute_config_id=compute_config_id,
            version="test_abc",
            canary_percent=canary_percent,
            rollout_strategy=rollout_strategy,
        )
    assert config == expected_service_config


@pytest.mark.parametrize("add_extra_field", [True, False])
def test_config_validation(add_extra_field, mock_auth_api_client):
    """
    This test checks that an error is thrown if extra/mispelled fields
    are added into the config file
    """
    service_controller = ServiceController()

    name = "test_service_name"
    description = "mock_description"
    build_id = "test_build_id"
    compute_config_id = "test_compute_config_id"
    project_id = "test_project_id"
    rollout_strategy = "ROLLOUT"
    extra_field_value = "*****"
    canary_percent = 100
    config_dict = {
        "name": name,
        "description": description,
        "ray_serve_config": {},
        "project_id": project_id,
        "build_id": build_id,
        "compute_config_id": compute_config_id,
        "version": "test_abc",
        "canary_percent": canary_percent,
        "rollout_strategy": rollout_strategy,
    }

    if add_extra_field:
        extra_field_value = "*****"
        config_dict["extra_field"] = extra_field_value

    mock_validate_successful_build = Mock()
    with patch(
        "builtins.open", mock_open(read_data=json.dumps(config_dict))
    ), patch.multiple("os.path", exists=Mock(return_value=True)), patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
    ):
        if add_extra_field:
            with pytest.raises(ValidationError):
                service_controller.generate_config_from_file(
                    "test_service_config_file", name=name, description=description,
                )
        else:
            service_controller.generate_config_from_file(
                "test_service_config_file", name=name, description=description,
            )


@pytest.mark.parametrize(
    ("working_dir"),
    [
        ("https://github.com/anyscale/docs_examples/archive/refs/heads/main.zip"),
        ("s3://already-a-uri/path"),
    ],
)
def test_generate_config_from_file_inside_workspace_no_working_dir_override(
    mock_auth_api_client, working_dir: str,
):
    """
    This test checks that if the service v2 is launched from the workspace with
    a remote working dir being specified, the method would inspect the remote url
    and not override it.
    """
    service_controller = ServiceController()

    name = "test_service_name"
    description = "mock_description"
    ray_serve_config = {
        "runtime_env": {"pip": ["requests"], "working_dir": working_dir,}
    }
    build_id = "test_build_id"
    compute_config_id = "test_compute_config_id"
    project_id = "test_project_id"
    rollout_strategy = "ROLLOUT"
    canary_percent = 100
    config_dict = {
        "name": name,
        "description": description,
        "ray_serve_config": ray_serve_config,
        "project_id": project_id,
        "build_id": build_id,
        "compute_config_id": compute_config_id,
        "version": "test_abc",
        "canary_percent": canary_percent,
        "rollout_strategy": rollout_strategy,
    }
    mock_validate_successful_build = Mock()

    workspace_id = "Test_workspace_id"
    with patch.dict(
        os.environ,
        {
            "ANYSCALE_EXPERIMENTAL_WORKSPACE_ID": workspace_id,
            "ANYSCALE_SESSION_ID": "test_session_id",
        },
    ), patch(
        "builtins.open", mock_open(read_data=json.dumps(config_dict))
    ), patch.multiple(
        "os.path", exists=Mock(return_value=True)
    ), patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
    ):
        config = service_controller.generate_config_from_file(
            "test_service_config_file",
        )

        expected_service_config = ServiceConfig(
            name=name,
            description=description,
            ray_serve_config=ray_serve_config,
            project_id=project_id,
            build_id=build_id,
            compute_config_id=compute_config_id,
            version="test_abc",
            canary_percent=canary_percent,
            rollout_strategy=rollout_strategy,
        )

    assert config == expected_service_config


@pytest.mark.parametrize("override_name", [None, "override_service_name"])
@pytest.mark.parametrize("override_version", [None, "override_version"])
@pytest.mark.parametrize(
    ("override_canary_percent", "first_get_ha_job_throws_error"),
    [[None, False], [0, False], [0, True], [40, False], [40, True]],
)
@pytest.mark.parametrize("override_rollout_strategy", [None, "IN_PLACE"])
def test_service_rollout(
    mock_auth_api_client,
    override_name: Optional[str],
    override_version: Optional[str],
    override_canary_percent: Optional[int],
    first_get_ha_job_throws_error: bool,
    override_rollout_strategy: Optional[str],
) -> None:
    """
    This tests that when we submit a Service config with ray_serve_config,
    we submit both the Service v1 and Service v2 configs.
    Please reference services_internal_api_models.py for more details.
    """
    service_controller = ServiceController()
    service_controller.api_client.apply_service_api_v2_services_v2_apply_put = Mock(
        return_value=Mock(
            result=(
                Mock(
                    versions=[
                        Mock(),
                        Mock(production_job_ids=["test_production_job_id"]),
                    ]
                )
            )
        )
    )
    if first_get_ha_job_throws_error:
        service_controller.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get = Mock(
            side_effect=[Exception("Error!"), Mock()]
        )

    name = "test_service_name"
    description = "mock_description"
    ray_serve_config = {"runtime_env": {"pip": ["requests"], "working_dir": "."}}
    build_id = "test_build_id"
    compute_config_id = "test_compute_config_id"
    project_id = "test_project_id"
    rollout_strategy = "ROLLOUT"
    service_config = {"access": {"use_bearer_token": True}}

    config_dict = {
        "name": name,
        "description": description,
        "ray_serve_config": ray_serve_config,
        "project_id": project_id,
        "build_id": build_id,
        "compute_config_id": compute_config_id,
        "version": "test_abc",
        "canary_percent": 100,
        "rollout_strategy": rollout_strategy,
        "config": service_config,
    }
    mock_validate_successful_build = Mock()
    mock_override_runtime_env_config = Mock(
        return_value={"pip": ["requests"], "working_dir": "."}
    )
    with patch(
        "builtins.open", mock_open(read_data=json.dumps(config_dict))
    ), patch.multiple("os.path", exists=Mock(return_value=True)), patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
    ), patch.multiple(
        "anyscale.controllers.service_controller",
        override_runtime_env_config=mock_override_runtime_env_config,
    ):
        service_controller.rollout(
            "test_ser8vice_config_file",
            name=override_name,
            version=override_version,
            canary_percent=override_canary_percent,
            rollout_strategy=override_rollout_strategy,
        )

    service_v2_config = ApplyProductionServiceV2Model(
        name=override_name if override_name else name,
        description=description,
        build_id=build_id,
        compute_config_id=compute_config_id,
        project_id=project_id,
        ray_serve_config=ray_serve_config,
        version=override_version if override_version else "test_abc",
        canary_percent=override_canary_percent
        if override_canary_percent is not None
        else 100,
        rollout_strategy=override_rollout_strategy
        if override_rollout_strategy
        else "ROLLOUT",
        config=service_config,
    )

    service_controller.api_client.apply_service_api_v2_services_v2_apply_put.assert_called_once_with(
        service_v2_config
    )


class TestGetServiceID:
    def test_invalid_inputs(self, mock_auth_api_client):
        service_controller = ServiceController()

        with pytest.raises(
            click.ClickException,
            match="Service ID, name, or config file must be specified.",
        ):
            service_controller.get_service_id()

        with pytest.raises(click.ClickException, match="Only one of"):
            service_controller.get_service_id(
                service_id="test-id", service_name="test-name"
            )

        with pytest.raises(click.ClickException, match="Only one of"):
            service_controller.get_service_id(
                service_id="test-id",
                service_config_file="test-config-file-path",
                project_id=None,
            )

        with pytest.raises(click.ClickException, match="Only one of"):
            service_controller.get_service_id(
                service_id="test-id",
                service_name="test-name",
                service_config_file="test-config-file-path",
            )

        with pytest.raises(click.ClickException, match="Only one of"):
            service_controller.get_service_id(
                service_name="test-name", service_config_file="test-config-file-path"
            )

        with pytest.raises(click.ClickException, match="not found"):
            service_controller.get_service_id(
                service_config_file="/fake/service/config/path"
            )

    def test_service_id_provided(self, mock_auth_api_client):
        """If the ID is provided, the same ID should always be returned and no API calls made."""
        service_controller = ServiceController()

        assert service_controller.get_service_id(service_id="test-id") == "test-id"
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_not_called()

    def test_name_provided_no_matching_service(self, mock_auth_api_client):
        """If a name is provided but no services match it, an exception should be raised."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[], metadata=Mock(next_paging_token=None),
        )

        with pytest.raises(
            click.ClickException,
            match=f"No service with name '{TEST_SERVICE_NAME}' was found",
        ):
            service_controller.get_service_id(service_name=TEST_SERVICE_NAME)

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10, creator_id=None, name=TEST_SERVICE_NAME, project_id=None,
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.reset_mock()

        # Test the config file codepath.
        mock_service_config = Mock()
        mock_service_config.configure_mock(name=TEST_SERVICE_NAME, project_id=None)
        service_controller._read_service_config_file = Mock(
            return_value=mock_service_config,
        )

        with pytest.raises(
            click.ClickException,
            match=f"No service with name '{TEST_SERVICE_NAME}' was found",
        ):
            service_controller.get_service_id(service_config_file="test-config-file")

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10, creator_id=None, name=TEST_SERVICE_NAME, project_id=None,
        )

    def test_name_provided_single_matching_service(self, mock_auth_api_client):
        """If a name is provided and one service matches it, its ID should be returned."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[Mock(id=TEST_SERVICE_ID)], metadata=Mock(next_paging_token=None),
        )

        assert (
            service_controller.get_service_id(service_name=TEST_SERVICE_NAME)
            == TEST_SERVICE_ID
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10, creator_id=None, name=TEST_SERVICE_NAME, project_id=None,
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.reset_mock()

        # Test the config file codepath.
        mock_service_config = Mock()
        mock_service_config.configure_mock(name=TEST_SERVICE_NAME, project_id=None)
        service_controller._read_service_config_file = Mock(
            return_value=mock_service_config,
        )
        assert (
            service_controller.get_service_id(service_config_file="test-config-file")
            == TEST_SERVICE_ID
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10, creator_id=None, name=TEST_SERVICE_NAME, project_id=None,
        )

    def test_name_provided_multiple_matching_services(self, mock_auth_api_client):
        """If a name is provided and multiple services match it, an exception should be raised."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[Mock(id="test-service-id-1"), Mock(id="test-service-id-2")],
            metadata=Mock(next_paging_token=None),
        )

        with pytest.raises(
            click.ClickException,
            match=f"There are multiple services with name '{TEST_SERVICE_NAME}'.",
        ):
            service_controller.get_service_id(service_name=TEST_SERVICE_NAME)

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10, creator_id=None, name=TEST_SERVICE_NAME, project_id=None,
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.reset_mock()

        mock_service_config = Mock()
        mock_service_config.configure_mock(name=TEST_SERVICE_NAME, project_id=None)
        service_controller._read_service_config_file = Mock(
            return_value=mock_service_config,
        )

        with pytest.raises(
            click.ClickException,
            match=f"There are multiple services with name '{TEST_SERVICE_NAME}'.",
        ):
            service_controller.get_service_id(service_config_file="test-config-file")

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10, creator_id=None, name=TEST_SERVICE_NAME, project_id=None,
        )

    def test_no_project_id(self, mock_auth_api_client):
        """If there is no project_id, the results should not be filtered based on project."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[Mock(id=TEST_SERVICE_ID)], metadata=Mock(next_paging_token=None),
        )

        mock_infer_project_id = Mock(return_value="mock_default_project_id")
        with patch.multiple(
            "anyscale.controllers.service_controller",
            infer_project_id=mock_infer_project_id,
        ):
            assert (
                service_controller.get_service_id(service_name=TEST_SERVICE_NAME)
                == TEST_SERVICE_ID
            )
            service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
                count=10, creator_id=None, name=TEST_SERVICE_NAME, project_id=None,
            )

    def test_project_id_in_service_config(self, mock_auth_api_client):
        """If the project_id is passed in the service config it should be used as a filter."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[Mock(id=TEST_SERVICE_ID)], metadata=Mock(next_paging_token=None),
        )

        mock_service_config = Mock()
        mock_service_config.configure_mock(
            name=TEST_SERVICE_NAME, project_id="test-project-id"
        )
        service_controller._read_service_config_file = Mock(
            return_value=mock_service_config,
        )

        mock_infer_project_id = Mock(return_value="mock_default_project_id")
        with patch.multiple(
            "anyscale.controllers.service_controller",
            infer_project_id=mock_infer_project_id,
        ):
            assert (
                service_controller.get_service_id(
                    service_config_file="test-config-file"
                )
                == TEST_SERVICE_ID
            )
            service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
                count=10,
                creator_id=None,
                name=TEST_SERVICE_NAME,
                project_id="test-project-id",
            )

    def test_override_project_id(self, mock_auth_api_client):
        """If the project_id is passed directly, the passed value should be used to filter."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[Mock(id=TEST_SERVICE_ID)], metadata=Mock(next_paging_token=None),
        )

        # Test overriding when getting service ID by name.
        assert (
            service_controller.get_service_id(
                service_name=TEST_SERVICE_NAME, project_id="overridden-project-id"
            )
            == TEST_SERVICE_ID
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10,
            creator_id=None,
            name=TEST_SERVICE_NAME,
            project_id="overridden-project-id",
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.reset_mock()

        # Test overriding when getting service ID by config file.
        mock_service_config = Mock()
        mock_service_config.configure_mock(
            name=TEST_SERVICE_NAME, project_id="test-project-id"
        )
        service_controller._read_service_config_file = Mock(
            return_value=mock_service_config,
        )

        mock_infer_project_id = Mock(return_value="mock_default_project_id")
        with patch.multiple(
            "anyscale.controllers.service_controller",
            infer_project_id=mock_infer_project_id,
        ):
            assert (
                service_controller.get_service_id(
                    service_config_file="test-config-file",
                    project_id="overridden-project-id",
                )
                == TEST_SERVICE_ID
            )

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10,
            creator_id=None,
            name=TEST_SERVICE_NAME,
            project_id="overridden-project-id",
        )


def test_terminate_service(mock_auth_api_client):
    """Test that terminate service"""
    service_controller = ServiceController()

    service_controller.terminate("service2_abc123")
    service_controller.api_client.terminate_service_api_v2_services_v2_service_id_terminate_post.assert_called_once_with(
        "service2_abc123",
    )


class TestListServices:
    def test_list_service_v2_with_id(self, mock_auth_api_client):
        service_controller = ServiceController()
        service_controller.list(service_id="service2_id")

        service_controller.api_client.get_service_api_v2_services_v2_service_id_get.assert_called_once_with(
            "service2_id",
        )

    @pytest.mark.parametrize("name", [None, "service_name"])
    @pytest.mark.parametrize("project_id", [None, "test_project_id"])
    @pytest.mark.parametrize("created_by_me", [True, False])
    def test_list_service(
        self,
        mock_auth_api_client,
        name: Optional[str],
        project_id: Optional[str],
        created_by_me: bool,
    ):
        service_controller = ServiceController()
        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[], metadata=Mock(next_paging_token=None),
        )
        if created_by_me:
            creator_id = "test_user_id"
            service_controller.api_client.get_user_info_api_v2_userinfo_get.return_value = Mock(
                result=Mock(id=creator_id)
            )
        else:
            creator_id = None

        service_controller.list(
            name=name, project_id=project_id, created_by_me=created_by_me
        )

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            creator_id=creator_id, name=name, project_id=project_id, count=10,
        )

    def test_list_service_with_pagination(
        self, mock_auth_api_client,
    ):
        service_controller = ServiceController()

        next_paging_token = "test"
        list_return_values = [
            DecoratedlistserviceapimodelListResponse(
                results=[],
                metadata=ListResponseMetadata(
                    total=11, next_paging_token=next_paging_token
                ),
            ),
            DecoratedlistserviceapimodelListResponse(
                results=[],
                metadata=ListResponseMetadata(total=11, next_paging_token=None),
            ),
        ]
        service_controller.api_client.list_services_api_v2_services_v2_get.side_effect = (
            list_return_values
        )

        service_controller.list()

        calls = [
            call(creator_id=None, name=None, project_id=None, count=10),
            call(
                creator_id=None,
                name=None,
                project_id=None,
                count=10,
                paging_token=next_paging_token,
            ),
        ]
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_has_calls(
            calls
        )


def test_rollback(mock_auth_api_client):
    """Test that v2 IDs go through v2 API, v1 is rejected."""
    service_controller = ServiceController()

    service_controller.rollback("service2_abc123")
    service_controller.api_client.rollback_service_api_v2_services_v2_service_id_rollback_post.assert_called_once_with(
        "service2_abc123",
    )


@pytest.mark.parametrize("init_runtime_env", [None, {}, {"working_dir": "."}])
def test_override_runtime_env_config_and_infer_project_id_v2(
    mock_auth_api_client, init_runtime_env: Collection[str],
):
    """Test helper functions in rollout function (service v2 path)
    Test that runtime_env is overwrriten from override_runtime_env_config
    Test the project_id is autopopulated from infer_project_id
    """
    config_dict = {
        "name": "mock_name",
        "description": "mock_description",
        "build_id": "mock_build_id",
        "compute_config_id": "mock_compute_config_id",
        "ray_serve_config": {
            "import_path": "mock_import_path",
            "runtime_env": init_runtime_env,
        },
    }

    assert "project_id" not in config_dict

    mock_project_id = "default"

    with patch.multiple(
        "anyscale.models.job_model", validate_successful_build=Mock(),
    ):
        service_config = ServiceConfig.parse_obj(config_dict)

    new_runtime_env = {"working_dir": "s3://bucket"}
    mock_override_config = Mock(return_value=new_runtime_env)

    mock_infer_project_id = Mock(return_value=mock_project_id)

    mock_generate_config_from_file = Mock(return_value=service_config)

    with patch.multiple(
        "anyscale.controllers.service_controller",
        infer_project_id=mock_infer_project_id,
        override_runtime_env_config=mock_override_config,
    ), patch.object(
        ServiceController, "generate_config_from_file", mock_generate_config_from_file
    ):
        service_controller = ServiceController()
        service_controller.api_client = Mock()
        service_controller.anyscale_api_client = Mock()
        service_controller.log = Mock()
        service_controller.api_client.apply_service_api_v2_services_v2_apply_put = (
            Mock()
        )

        service_controller.rollout(service_config)

        mock_override_config.assert_called_once_with(
            anyscale_api_client=service_controller.anyscale_api_client,
            api_client=service_controller.api_client,
            compute_config_id=config_dict["compute_config_id"],
            runtime_env=init_runtime_env,
            workload_type=Workload.SERVICES,
            log=service_controller.log,
        )

        expected_ray_serve_config = {
            "import_path": "mock_import_path",
            "runtime_env": new_runtime_env,
        }

        service_controller.api_client.apply_service_api_v2_services_v2_apply_put.assert_called_once_with(
            ApplyProductionServiceV2Model(
                name=config_dict["name"],
                description=config_dict["description"],
                project_id=mock_project_id,
                canary_percent=None,
                ray_serve_config=expected_ray_serve_config,
                build_id=config_dict["build_id"],
                compute_config_id=service_config.compute_config_id,
                rollout_strategy=service_config.rollout_strategy,
            )
        )


def test_overwrite_runtime_env_in_v2_ray_serve_config(mock_auth_api_client):
    """Test that runtime env is overwritten in v2 single-app configs as expected.
    Delegates the runtime env overwriting logic to override_runtime_env_config(), simply
    makes sure that override_runtime_env_config is called with the correct arguments.
    """
    service_controller = ServiceController()

    # Runtime environments
    original_runtime_env = {
        "pip": ["requests"],
        "working_dir": ".",
        "upload_path": "s3://already-a-uri/path",
    }
    rewritten_runtime_env = {
        "pip": ["requests"],
        "working_dir": "s3://bucket",
    }

    # Service configs
    with patch.multiple(
        "anyscale.models.job_model", validate_successful_build=Mock(),
    ):
        service_config = ServiceConfig.parse_obj(
            {
                "name": "mock_name",
                "description": "mock_description",
                "build_id": "mock_build_id",
                "compute_config_id": "mock_compute_config_id",
                "ray_serve_config": {
                    "import_path": "a.b.c",
                    "runtime_env": original_runtime_env,
                },
            }
        )
        expected_service_config = deepcopy(service_config)
        expected_service_config.ray_serve_config["runtime_env"] = rewritten_runtime_env

    # Perform test
    override_runtime_env_config_mock = Mock(return_value=rewritten_runtime_env)
    with patch.multiple(
        "anyscale.controllers.service_controller",
        override_runtime_env_config=override_runtime_env_config_mock,
    ):
        service_controller._overwrite_runtime_env_in_v2_ray_serve_config(service_config)

    calls = [
        call(
            runtime_env=original_runtime_env,
            anyscale_api_client=ANY,
            api_client=ANY,
            workload_type=Workload.SERVICES,
            compute_config_id="mock_compute_config_id",
            log=ANY,
        )
    ]
    assert service_config == expected_service_config
    override_runtime_env_config_mock.assert_has_calls(calls)


def test_overwrite_runtime_env_in_v2_ray_serve_config_multi_app(mock_auth_api_client):
    """Test that runtime env for each app is overwritten in v2 multi-app configs.
    Delegates the runtime env overwriting logic to override_runtime_env_config(), simply
    makes sure that override_runtime_env_config is called with the correct arguments.
    """
    service_controller = ServiceController()

    # Runtime environments
    original_runtime_env1 = {
        "pip": ["requests"],
        "working_dir": ".",
        "upload_path": "s3://already-a-uri/path",
    }
    original_runtime_env2 = None
    rewritten_runtime_env = {
        "pip": ["requests"],
        "working_dir": "s3://bucket",
    }

    # Service configs
    with patch.multiple(
        "anyscale.models.job_model", validate_successful_build=Mock(),
    ):
        service_config = ServiceConfig.parse_obj(
            {
                "name": "mock_name",
                "description": "mock_description",
                "build_id": "mock_build_id",
                "compute_config_id": "mock_compute_config_id",
                "ray_serve_config": {
                    "applications": [
                        {"import_path": "a.b", "runtime_env": original_runtime_env1},
                        {"import_path": "c.d"},
                    ]
                },
            }
        )
        expected_service_config = deepcopy(service_config)
        for app in expected_service_config.ray_serve_config["applications"]:
            app["runtime_env"] = rewritten_runtime_env

    # Perform test
    override_runtime_env_config_mock = Mock(return_value=rewritten_runtime_env)
    with patch.multiple(
        "anyscale.controllers.service_controller",
        override_runtime_env_config=override_runtime_env_config_mock,
    ):
        service_controller._overwrite_runtime_env_in_v2_ray_serve_config(service_config)

    calls = [
        call(
            runtime_env=original_runtime_env1,
            anyscale_api_client=ANY,
            api_client=ANY,
            workload_type=Workload.SERVICES,
            compute_config_id="mock_compute_config_id",
            log=ANY,
        ),
        call(
            runtime_env=original_runtime_env2,
            anyscale_api_client=ANY,
            api_client=ANY,
            workload_type=Workload.SERVICES,
            compute_config_id="mock_compute_config_id",
            log=ANY,
        ),
    ]
    assert service_config == expected_service_config
    override_runtime_env_config_mock.assert_has_calls(calls)



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class CreateFineTuningHyperparameters(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'n_epochs': 'int',
        'context_length': 'int'
    }

    attribute_map = {
        'n_epochs': 'n_epochs',
        'context_length': 'context_length'
    }

    def __init__(self, n_epochs=None, context_length=None, local_vars_configuration=None):  # noqa: E501
        """CreateFineTuningHyperparameters - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._n_epochs = None
        self._context_length = None
        self.discriminator = None

        if n_epochs is not None:
            self.n_epochs = n_epochs
        if context_length is not None:
            self.context_length = context_length

    @property
    def n_epochs(self):
        """Gets the n_epochs of this CreateFineTuningHyperparameters.  # noqa: E501

        Number of epochs to train for. Auto-calculated based on the training file size if not specified  # noqa: E501

        :return: The n_epochs of this CreateFineTuningHyperparameters.  # noqa: E501
        :rtype: int
        """
        return self._n_epochs

    @n_epochs.setter
    def n_epochs(self, n_epochs):
        """Sets the n_epochs of this CreateFineTuningHyperparameters.

        Number of epochs to train for. Auto-calculated based on the training file size if not specified  # noqa: E501

        :param n_epochs: The n_epochs of this CreateFineTuningHyperparameters.  # noqa: E501
        :type: int
        """

        self._n_epochs = n_epochs

    @property
    def context_length(self):
        """Gets the context_length of this CreateFineTuningHyperparameters.  # noqa: E501

        Maximum number of tokens to be considered for each training example. Auto-calculated based on the P95 sequence length of all training examples if not specified  # noqa: E501

        :return: The context_length of this CreateFineTuningHyperparameters.  # noqa: E501
        :rtype: int
        """
        return self._context_length

    @context_length.setter
    def context_length(self, context_length):
        """Sets the context_length of this CreateFineTuningHyperparameters.

        Maximum number of tokens to be considered for each training example. Auto-calculated based on the P95 sequence length of all training examples if not specified  # noqa: E501

        :param context_length: The context_length of this CreateFineTuningHyperparameters.  # noqa: E501
        :type: int
        """
        allowed_values = [512, 1024, 2048, 4096]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and context_length not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `context_length` ({0}), must be one of {1}"  # noqa: E501
                .format(context_length, allowed_values)
            )

        self._context_length = context_length

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CreateFineTuningHyperparameters):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CreateFineTuningHyperparameters):
            return True

        return self.to_dict() != other.to_dict()



from abc import ABC, abstractmethod
import asyncio
import contextlib
from datetime import datetime, timedelta
import json
import os
import random
import string
import time
import traceback
from typing import Any, Callable, cast, Dict, List, Optional, Tuple, Union

import click
from pydantic import BaseModel
import requests
from requests.adapters import HTTPAdapter
import tabulate
from urllib3 import Retry
import yaml

from anyscale.api_utils.exceptions.log_retrieval_errors import (
    ExpectedLogRetrievalError,
    wrap_as_unsupported_log_retrieval_method_error,
    wrap_job_run_log_not_retrievable_on_active_cluster_error,
)
from anyscale.api_utils.job_logs_util import (
    _get_job_logs_from_storage_bucket,
    _get_job_logs_from_storage_bucket_streaming,
    _get_logs_from_active_job_run,
)
from anyscale.api_utils.logs_util import _stream_log_from_ray_websocket
from anyscale.cli_logger import LogsLogger
from anyscale.client.openapi_client import (
    ComputeTemplate,
    ProductionJob,
    ProductionJobConfig,
)
from anyscale.client.openapi_client.models.create_internal_production_job import (
    CreateInternalProductionJob,
)
from anyscale.client.openapi_client.models.decorated_production_job import (
    DecoratedProductionJob,
)
from anyscale.client.openapi_client.models.ha_job_states import HaJobStates
from anyscale.controllers.base_controller import BaseController
from anyscale.models.job_model import JobConfig
from anyscale.project import infer_project_id
from anyscale.sdk.anyscale_client.models.job import Job
from anyscale.sdk.anyscale_client.models.jobs_query import JobsQuery
from anyscale.sdk.anyscale_client.models.jobs_sort_field import JobsSortField
from anyscale.sdk.anyscale_client.models.page_query import PageQuery
from anyscale.sdk.anyscale_client.models.sort_by_clause_jobs_sort_field import (
    SortByClauseJobsSortField,
)
from anyscale.sdk.anyscale_client.models.sort_order import SortOrder
from anyscale.util import (
    get_endpoint,
    handle_and_log_lines_in_click_exception,
    is_anyscale_workspace,
    poll,
    populate_dict_with_workspace_config_if_exists,
    validate_job_config_dict,
)
from anyscale.utils.connect_helpers import search_entities
from anyscale.utils.runtime_env import override_runtime_env_config
from anyscale.utils.workload_types import Workload


log = LogsLogger()

_TERMINAL_STATES = {
    HaJobStates.SUCCESS,
    HaJobStates.TERMINATED,
    HaJobStates.BROKEN,
    HaJobStates.OUT_OF_RETRIES,
}
_PENDING_STATES = {
    HaJobStates.PENDING,
    HaJobStates.AWAITING_CLUSTER_START,
    HaJobStates.ERRORED,
    HaJobStates.RESTARTING,
}


class MiniJobRun(BaseModel):
    last_job_run_id: str
    job_state: str
    error: Optional[str]


class LogProvider(ABC):
    @abstractmethod
    def __init__(self, dns_name: str, anyscale_token: str, query: str) -> None:
        pass

    @abstractmethod
    def query(
        self, start_timestamp_ns: int, end_timestamp_ns: int
    ) -> List[Tuple[int, str]]:
        """Query logs for a time range, return list of (timestamp_ns, log_line)."""

    @abstractmethod
    def close(self) -> None:
        """Close any connection this log provider uses"""

    def __enter__(self, *args, **kwargs):  # noqa: B027
        pass

    def __exit__(self, *args, **kwargs):
        self.close()


class LokiLogProvider(LogProvider):
    def __init__(self, dns_name: str, anyscale_token: str, query: str) -> None:
        # https://grafana.com/docs/loki/latest/api/#get-lokiapiv1query_range
        self.url = f"https://{dns_name}/loki/api/v1/query_range"
        self.params: Dict[str, Union[str, int]] = {
            "token": anyscale_token,
            "query": query,
            "limit": 1000,
            "start": 0,
            "end": int(time.time() * 1e9),
            "direction": "forward",
        }

    def __enter__(self, *args, **kwargs):
        self.session = requests.Session()
        self.session.mount(
            "https://",
            HTTPAdapter(
                max_retries=Retry(total=10, backoff_factor=0.1, allowed_methods=["GET"])
            ),
        )
        return self

    def close(self) -> None:
        self.session.close()

    def __exit__(self, *args, **kwargs):
        self.close()

    def query(
        self, start_timestamp_ns: int, end_timestamp_ns: int
    ) -> List[Tuple[int, str]]:
        if start_timestamp_ns == 0:
            # 10^9 nanoseconds in a second
            start_timestamp_ns = int(
                (datetime.now() - timedelta(days=7)).timestamp() * (10 ** 9)
            )
        self.params["start"] = start_timestamp_ns
        self.params["end"] = end_timestamp_ns

        resp = self.session.get(self.url, params=self.params)
        resp.raise_for_status()

        resp_data = resp.json()
        if resp_data["status"] != "success":
            raise click.ClickException("Querying Anyscale log endpoint failed.")
        if resp_data["data"]["resultType"] != "streams":
            raise click.ClickException("Invalid metrics type.")

        lines = []
        for log_chunk in resp_data["data"]["result"]:
            # TODO(simon): potentially annotate the log with the log_chunk["stream"] metadata
            for timestamp, line in log_chunk["values"]:
                lines.append((int(timestamp), line))

        return lines


class JobController(BaseController):
    def __init__(
        self,
        log: Optional[LogsLogger] = None,
        initialize_auth_api_client: bool = True,
        raise_structured_exception: bool = False,
    ):
        if log is None:
            log = LogsLogger()

        super().__init__(initialize_auth_api_client, raise_structured_exception)
        self.log = log
        self.log.open_block("Output")
        self._cluster_journal_events_start_line = 0

    def submit(
        self,
        job_config_file: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_entrypoint_cmd: Optional[bool] = False,
        entrypoint: Optional[List[str]] = None,
    ) -> str:
        entrypoint = entrypoint or []
        workspace_id = os.environ.get("ANYSCALE_EXPERIMENTAL_WORKSPACE_ID", None)
        if is_anyscale_workspace() and is_entrypoint_cmd:
            entrypoint = [job_config_file, *entrypoint]
            config = self.generate_config_from_entrypoint(
                entrypoint, name, description, workspace_id
            )
            id = self.submit_from_config(config)  # noqa: A001
        elif len(entrypoint) == 0:
            # Assume that job_config_file is a file and submit it.
            config = self.generate_config_from_file(
                job_config_file,
                name=name,
                description=description,
                workspace_id=workspace_id,
            )
            id = self.submit_from_config(config)  # noqa: A001
        elif len(entrypoint) != 0:
            msg = (
                "Within an Anyscale Workspace, `anyscale job submit` takes either a file, or a command. To submit a command, use `anyscale job submit -- my command`."
                if is_anyscale_workspace()
                else "`anyscale job submit` takes one argument, a YAML file configuration. Please use `anyscale job submit my_file`."
            )
            raise click.ClickException(msg)
        return id

    def generate_config_from_entrypoint(
        self,
        entrypoint: List[str],
        name: Optional[str],
        description: Optional[str],
        workspace_id: Optional[str] = None,
    ) -> JobConfig:
        config_dict = {
            "entrypoint": " ".join(entrypoint),
            "name": name,
            "description": description,
            "workspace_id": workspace_id,
        }
        config_dict = populate_dict_with_workspace_config_if_exists(
            config_dict, self.anyscale_api_client
        )
        job_config = JobConfig.parse_obj(config_dict)
        return job_config

    def generate_config_from_file(
        self,
        job_config_file: str,
        name: Optional[str],
        description: Optional[str],
        workspace_id: Optional[str] = None,
    ) -> JobConfig:
        config_dict = self._load_config_dict_from_file(job_config_file)
        config_dict["workspace_id"] = workspace_id
        validate_job_config_dict(config_dict, self.api_client)
        config_dict = populate_dict_with_workspace_config_if_exists(
            config_dict, self.anyscale_api_client
        )
        job_config = JobConfig.parse_obj(config_dict)

        if name:
            job_config.name = name

        if description:
            job_config.description = description

        return job_config

    def submit_from_config(self, job_config: JobConfig):
        # If project id is not specified, try to infer it
        project_id = infer_project_id(
            self.anyscale_api_client,
            self.api_client,
            self.log,
            project_id=job_config.project_id,
            cluster_compute_id=job_config.compute_config_id,
            cluster_compute=job_config.compute_config,
            cloud=job_config.cloud,
            cluster_compute_version=job_config.compute_config_version,
        )

        job_config.runtime_env = override_runtime_env_config(
            runtime_env=job_config.runtime_env,
            anyscale_api_client=self.anyscale_api_client,
            api_client=self.api_client,
            workload_type=Workload.JOBS,
            compute_config_id=job_config.compute_config_id,
            log=self.log,
        )
        config_object = ProductionJobConfig(
            entrypoint=job_config.entrypoint,
            runtime_env=job_config.runtime_env,
            build_id=job_config.build_id,
            compute_config_id=job_config.compute_config_id,
            max_retries=job_config.max_retries,
        )

        job = self.api_client.create_job_api_v2_decorated_ha_jobs_create_post(
            CreateInternalProductionJob(
                name=job_config.name or self._generate_random_job_name(),
                description=job_config.description or "Job submitted from CLI",
                project_id=project_id,
                workspace_id=job_config.workspace_id,
                config=config_object,
            )
        ).result
        self.log.info(
            f"Maximum uptime is {self._get_maximum_uptime_output(job)} for clusters launched by this job."
        )
        self.log.info(
            f"Job {job.id} has been successfully submitted. Current state of job: {job.state.current_state}."
        )
        self.log.info(
            f"Query the status of the job with `anyscale job list --job-id {job.id}`."
        )
        self.log.info(
            f"Get the logs for the job with `anyscale job logs --job-id {job.id} --follow`."
        )
        self.log.info(f'View the job in the UI at {get_endpoint(f"/jobs/{job.id}")}')
        return job.id

    def _get_maximum_uptime_output(self, job: ProductionJob) -> str:
        compute_config: ComputeTemplate = self.api_client.get_compute_template_api_v2_compute_templates_template_id_get(
            job.config.compute_config_id
        ).result
        maximum_uptime_minutes = compute_config.config.maximum_uptime_minutes
        if maximum_uptime_minutes and maximum_uptime_minutes > 0:
            return f"set to {maximum_uptime_minutes} minutes"
        return "disabled"

    def _load_config_dict_from_file(self, job_config_file: str) -> Dict[str, Any]:
        if not os.path.exists(job_config_file):
            raise click.ClickException(f"Config file {job_config_file} not found.")

        with open(job_config_file) as f:
            config_dict = yaml.safe_load(f)
        return config_dict

    def wait(
        self,
        job_name: Optional[str] = None,
        job_id: Optional[str] = None,
        target_state: str = HaJobStates.SUCCESS,
        interval_secs: float = 10.0,
        timeout_secs=None,
    ):
        if target_state not in HaJobStates.allowable_values:
            raise click.ClickException(
                f"`{target_state}` is not a valid Job state. Allowed states are {HaJobStates.allowable_values}"
            )
        job_id_or_name = job_name or job_id
        job_id = cast(str, self._resolve_job_object(job_id, job_name).id,)
        with self.log.spinner(f"Waiting for Job `{job_id_or_name}`...") as spinner:
            state = None
            for _ in poll(interval_secs=interval_secs, timeout_secs=timeout_secs):
                state = self._get_job_state(job_id)
                spinner.text = f"Job `{job_id_or_name}` is in state `{state}`. Waiting to reach state `{target_state}`."

                if state == target_state:
                    spinner.succeed(
                        f"Job `{job_id_or_name}` reached state `{target_state}`."
                    )
                    return job_id

                if state in _TERMINAL_STATES:
                    msg = f"Job `{job_id_or_name}` reached terminal state `{state}`, and will not reach `{target_state}`."
                    spinner.fail(msg)
                    raise click.ClickException(msg)

            msg = f"Timed out after waiting for {timeout_secs} seconds. The current state is {state}."
            spinner.fail(msg)
            raise click.ClickException(msg)

    def _get_job_state(self, job_id: str):
        job = self.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
            job_id
        ).result
        return job.state.current_state

    def list(  # noqa: PLR0913
        self,
        include_all_users: bool,
        name: Optional[str],
        job_id: Optional[str],
        project_id: Optional[str],
        include_archived: bool,
        max_items: int,
    ) -> None:
        """
        This function will list jobs.
        """
        print(f'View your Jobs in the UI at {get_endpoint("/jobs")}')

        jobs_list = []
        if job_id:
            job = self.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
                job_id
            ).result
            if job.is_service:
                # `job_id` belongs to a service, but this function should list jobs.
                raise click.ClickException(
                    f"ID {job_id} belongs to a Anyscale service. Please get information about "
                    f"this service with `anyscale service list --service-id {job_id}`."
                )

            output_map = {
                "Name": job.name,
                "Id": job.id,
                "Cost (dollars)": job.cost_dollars,
                "Project name": job.project.name,
                "Cluster name": job.last_job_run.cluster.name
                if job.last_job_run and job.last_job_run.cluster
                else None,
                "Current state": job.state.current_state,
                "Creator": job.creator.email,
                "Entrypoint": job.config.entrypoint
                if len(job.config.entrypoint) < 100
                else job.config.entrypoint[:100] + " ...",
            }
            output_str = "\n".join(
                [f"\t{key}: {output_map[key]}" for key in output_map]
            )
            print(output_str)
            return
        else:
            if not include_all_users:
                creator_id = (
                    self.api_client.get_user_info_api_v2_userinfo_get().result.id
                )
            else:
                creator_id = None
            resp = self.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get(
                project_id=project_id,
                name=name,
                creator_id=creator_id,
                type_filter="BATCH_JOB",
                archive_status="ALL" if include_archived else "NOT_ARCHIVED",
                count=10,
            )
            jobs_list.extend(resp.results)
            paging_token = resp.metadata.next_paging_token
            has_more = (paging_token is not None) and (len(jobs_list) < max_items)
            while has_more:
                resp = self.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get(
                    project_id=project_id,
                    name=name,
                    creator_id=creator_id,
                    type_filter="BATCH_JOB",
                    archive_status="ALL" if include_archived else "NOT_ARCHIVED",
                    count=10,
                    paging_token=paging_token,
                )
                jobs_list.extend(resp.results)
                paging_token = resp.metadata.next_paging_token
                has_more = (paging_token is not None) and (len(jobs_list) < max_items)
            jobs_list = jobs_list[:max_items]

        jobs_table = [
            [
                job.name,
                job.id,
                job.cost_dollars,
                job.project.name,
                job.last_job_run.cluster.name
                if job.last_job_run and job.last_job_run.cluster
                else None,
                job.state.current_state,
                job.creator.email,
                job.config.entrypoint
                if len(job.config.entrypoint) < 50
                else job.config.entrypoint[:50] + " ...",
            ]
            for job in jobs_list
        ]

        table = tabulate.tabulate(
            jobs_table,
            headers=[
                "NAME",
                "ID",
                "COST",
                "PROJECT NAME",
                "CLUSTER NAME",
                "CURRENT STATE",
                "CREATOR",
                "ENTRYPOINT",
            ],
            tablefmt="plain",
        )
        print(f"JOBS:\n{table}")

    def archive(self, job_id: Optional[str], job_name: Optional[str]) -> None:
        """
        This function will archive jobs.
        """
        job_resp: DecoratedProductionJob = self._resolve_job_object(job_id, job_name)
        self.api_client.archive_job_api_v2_decorated_ha_jobs_production_job_id_archive_post(
            job_resp.id
        )
        self.log.info(f"Job {job_resp.id} is successfully archived.")

    def terminate(self, job_id: Optional[str], job_name: Optional[str]) -> None:
        """
        This function will terminate jobs.
        """
        job_resp: DecoratedProductionJob = self._resolve_job_object(job_id, job_name)
        job = self.api_client.terminate_job_api_v2_decorated_ha_jobs_production_job_id_terminate_post(
            job_resp.id
        ).result
        self.log.info(f"Job {job.id} has begun terminating...")
        self.log.info(
            f" Current state of Job: {job.state.current_state}. Goal state of Job: {job.state.goal_state}"
        )
        self.log.info(
            f"Query the status of the Job with `anyscale job list --job-id {job.id}`."
        )

    def _resolve_job_object(
        self, job_id: Optional[str], job_name: Optional[str]
    ) -> DecoratedProductionJob:
        """Given job_id or job_name, retrieve decorated ha job spec"""
        if job_id is None and job_name is None:
            raise click.ClickException(
                "Either `--id` or `--name` must be passed in for Job."
            )
        if job_id:
            return self._get_job(job_id)

        jobs_list_resp: List[
            DecoratedProductionJob
        ] = self.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get(
            name=job_name, type_filter="BATCH_JOB",
        ).results
        if len(jobs_list_resp) == 0:
            raise click.ClickException(
                f"No Job found with name {job_name}. Please either pass `--id` or list the "
                f"available Jobs with `anyscale job list`."
            )
        if len(jobs_list_resp) > 1:
            raise click.ClickException(
                f"Multiple Jobs found with name {job_name}. Please specify the `--id` instead."
            )
        return jobs_list_resp[0]

    def _get_job(self, job_id: str) -> DecoratedProductionJob:
        job_object = self.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
            production_job_id=job_id
        ).result
        return job_object

    def _get_formatted_latest_job_run(
        self, job: DecoratedProductionJob
    ) -> Optional[MiniJobRun]:
        job_state = job.state.current_state
        last_job_run_id = job.last_job_run_id
        if job_state in _TERMINAL_STATES and last_job_run_id is None:
            raise click.ClickException(
                f"Can't find latest job run for {job_state} job."
            )
        if not last_job_run_id:
            return None
        return MiniJobRun(
            last_job_run_id=last_job_run_id, job_state=job_state, error=job.state.error
        )

    def _wait_for_a_job_run(
        self, job_id: Optional[str], job_name: Optional[str]
    ) -> DecoratedProductionJob:
        """Waits until the job has a job run and is in a non-pending state
        (encountered when initially submitting).

        Returns the job."""
        start = time.monotonic()
        job = self._resolve_job_object(job_id, job_name)
        last_job_run_id: str = job.last_job_run_id
        job_state = job.state.current_state
        with self.log.spinner("Waiting for job run...") as spinner:
            while last_job_run_id is None or job_state in _PENDING_STATES:
                if job_state in _TERMINAL_STATES:
                    raise click.ClickException(
                        f"Can't find latest job run for {job_state} job."
                    )
                if self._cluster_journal_events_start_line == 0:
                    spinner.text = (
                        f"Waiting for a job run, current state is {job_state}..."
                    )
                time.sleep(max(0, 3 - (time.monotonic() - start)))
                start = time.monotonic()
                job = self._get_job(job.id)
                last_job_run_id = job.last_job_run_id
                if job.state.cluster_id:
                    # Once we have a cluster id, we start printing the cluster journal event logs with each loop.
                    cluster_journal_events = self._api_client.get_startup_logs_api_v2_sessions_session_id_startup_logs_get(
                        job.state.cluster_id,
                        start_line=self._cluster_journal_events_start_line,
                        end_line=10000000,
                    ).result
                    lines = cluster_journal_events.lines.splitlines()
                    if len(lines):
                        spinner.succeed("Job run found. Cluster launching...")
                    self._cluster_journal_events_start_line = (
                        cluster_journal_events.start_line
                        + cluster_journal_events.num_lines
                    )

                    for line in lines:
                        self.log.info(line)

                job_state = job.state.current_state
        return job

    def _print_logs_for_active_job_run(
        self, job_run_id: str, should_follow: bool
    ) -> None:
        """
        Prints logs from the active cluster the job is run on.
        """
        with handle_and_log_lines_in_click_exception(log_callback=self.log.error):
            if should_follow:
                # Use websockets (to continuously poll logs)
                with wrap_as_unsupported_log_retrieval_method_error():
                    websocket_url: str = self.api_client.get_job_logs_stream_api_v2_logs_job_logs_stream_job_id_get(
                        job_id=job_run_id
                    ).result.stream_url

                with wrap_job_run_log_not_retrievable_on_active_cluster_error(
                    job_run_id, raise_connection_issue_as_cli_error=True
                ):
                    # (run_until_complete() is used here b/c the KeyboardInterrupt exception is
                    # not propogated all the way up (and likely being handled by internals of asyncio.run()))
                    asyncio.get_event_loop().run_until_complete(
                        _stream_log_from_ray_websocket(
                            websocket_url,
                            log_callback=self.log.log,
                            remove_escape_chars=False,
                        )
                    )
            else:
                # Make a one-time request to get logs
                logs = asyncio.run(
                    _get_logs_from_active_job_run(
                        self.anyscale_api_client,
                        job_run_id,
                        raise_connection_issue_as_cli_error=True,
                        remove_escape_chars=False,
                    )
                )
                for line in logs.splitlines():
                    self.log.log(line)

    def _print_logs_for_inactive_job_run(self, job_run_id: str) -> None:
        """Retrieves logs from the cloud storage bucket for a job run in terminal state"""
        logs = ""
        with self.log.spinner("Retrieving logs..."):
            # For jobs on a cluster that is terminating, the logs are in-flight between cluster and storage bucket
            # of the cloud. In that case, wait for a bit and try again.
            for _ in poll(interval_secs=5, max_iter=10):
                logs = asyncio.run(
                    _get_job_logs_from_storage_bucket(
                        self.anyscale_api_client,
                        job_run_id=job_run_id,
                        remove_escape_chars=False,
                    )
                )
                if logs:
                    break
        if not logs:
            self.log.error("Logs are not available for this job run.")
            return
        for line in logs.splitlines():
            self.log.log(line)

    def _print_logs_using_streaming_job_logs(
        self, job_run_id: str, follow: bool = False
    ) -> None:
        """Retrieves logs from the streaming job logs folder in S3/GCS"""
        logs = ""
        next_page_token = None
        # Keep track of if any logs have been printed yet,
        # If not, we assume the cluster recently launched and we
        # give more leeway for the logs to be available.
        have_jobs_started_producing_logs = False
        with self.log.spinner("Retrieving logs..."):
            (
                logs,
                next_page_token,
                cluster_journal_events_start_line,
                job_finished,
            ) = asyncio.run(
                _get_job_logs_from_storage_bucket_streaming(
                    self._api_client,
                    self.log,
                    job_run_id=job_run_id,
                    remove_escape_chars=False,
                    next_page_token=next_page_token,
                    cluster_journal_events_start_line=self._cluster_journal_events_start_line,
                )
            )

        if len(logs):
            have_jobs_started_producing_logs = True

        for line in logs.splitlines():
            self.log.log(line)

        # Continuously poll for logs.
        terminal_state_count = 0
        while follow:
            start = time.monotonic()
            if job_finished:
                # Let's wait before terminating the loop to give the logs a chance to catch up.
                terminal_state_count += 1
                # Once jobs produce logs, we know we do not have to wait as long for those logs
                # to be uploaded to S3 since the cluster is already running.
                max_iterations = 4 if have_jobs_started_producing_logs else 6
                if terminal_state_count >= max_iterations:
                    self.log.info("Job run reached terminal state.")
                    return

            (
                logs,
                next_page_token,
                cluster_journal_events_start_line,
                job_finished,
            ) = asyncio.run(
                _get_job_logs_from_storage_bucket_streaming(
                    self._api_client,
                    self.log,
                    job_run_id=job_run_id,
                    remove_escape_chars=False,
                    next_page_token=next_page_token,
                    cluster_journal_events_start_line=cluster_journal_events_start_line,
                )
            )

            for line in logs.splitlines():
                self.log.log(line)

            # Wait at least 5 seconds between iterations.
            time.sleep(max(0, 5 - (time.monotonic() - start)))

    def logs(
        self,
        job_id: Optional[str] = None,
        job_name: Optional[str] = None,
        should_follow: bool = False,
        all_attempts: bool = False,
    ) -> None:
        """Uses fallback logic to print logs for a (production) job to terminal output.

        * If the job is running, it will retrieve logs from the cluster the job is run on.
        * If the job is not running, it will retrieve logs from Anyscale's backend cache.
        * If the logs no longer live in the cache, or the cloud the job is on is using Anyscale V2,
            it falls back to retrieving logs from the storage bucket of the cloud.
        """
        # This logic will be simplified once the migration to Anyscale V2 is complete.
        job = self._wait_for_a_job_run(job_id, job_name)
        is_streaming_job_logs_on = self._api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get(
            "ray-job-log-streaming"
        ).result.is_on
        last_job_run_id: str = job.last_job_run_id
        job_state: HaJobStates = job.state.current_state
        job_run_ids: List[str] = (
            [last_job_run_id]
            if not all_attempts
            else [job_run.id for job_run in self._get_all_job_runs(job.id)]
        )  # type: ignore
        for job_run_id in job_run_ids:
            self.log.open_block(
                job_run_id, f"Job Run Id: {job_run_id}", auto_close=True,
            )
            if is_streaming_job_logs_on:
                # New path for retrieving logs. Works regardless if cluster is running or dead.
                self._print_logs_using_streaming_job_logs(
                    job_run_id, follow=should_follow and job_run_id == last_job_run_id
                )
            else:
                # 1. Last job run, and 2. job is RUNNING implies (==) job run is RUNNING
                is_last_job_run_and_active = (
                    job_run_id == last_job_run_id and job_state == HaJobStates.RUNNING
                )

                get_logs_fns: List[Callable[[], None]] = [
                    *(
                        # If the job run is RUNNING, retrieve logs from the active cluster
                        []
                        if not is_last_job_run_and_active
                        else [
                            lambda: self._print_logs_for_active_job_run(
                                job_run_id, should_follow  # noqa: B023
                            )
                        ]
                    ),
                    # If the previous retrieval method failed or the job is not running
                    # retrieve logs from Anyscale's cache
                    lambda: self._print_logs_for_inactive_job_run(
                        job_run_id  # noqa: B023
                    ),
                ]
                for fn in get_logs_fns:
                    with contextlib.suppress(ExpectedLogRetrievalError):
                        fn()
                        # If `fn` didn't raise `ExpectedLogRetrievalError`, we assume logs were successfully printed.
                        # In this case, break out of this loop and log for the next `job_run_id`
                        break

    def retrieve_output(
        self, job_name: Optional[str] = None, job_id: Optional[str] = None,
    ) -> None:
        if not job_id:
            job = self._resolve_job_object(job_id, job_name)
            job_id = job.id

        try:
            output = self.api_client.find_structured_job_output_api_v2_structured_outputs_ha_job_output_ha_job_id_get(
                ha_job_id=job_id
            )
            self.log.log(json.dumps(output.result.output, indent=2))
        except Exception as e:  # noqa: BLE001
            traceback.print_exc()
            # Swallow if pretty print logic fails
            with contextlib.suppress(Exception):
                j = json.loads(e.body)  # type: ignore
                detail = j.get("error", {}).get("detail")
                self.log.error(detail)

    def _generate_random_job_name(self) -> str:
        """Generates a random job name
        Format:
            cli-job-{10 random characters}
        """
        random_chars = "".join(random.choices(string.ascii_letters, k=10)).lower()
        return "cli-job-" + random_chars

    def _get_all_job_runs(self, job_id: str) -> List[Job]:
        """Returns all job runs for a given job id.
        Returned in ascending order by creation time."""
        job_runs: List[Job] = search_entities(
            self.anyscale_api_client.search_jobs,
            JobsQuery(
                ha_job_id=job_id,
                show_ray_client_runs_only=False,
                sort_by_clauses=[
                    SortByClauseJobsSortField(
                        sort_field=JobsSortField.CREATED_AT, sort_order=SortOrder.ASC,
                    )
                ],
                paging=PageQuery(),
            ),
        )
        return job_runs



from anyscale.client.openapi_client.models.apply_production_service_v2_model import (
    ApplyProductionServiceV2Model,
)
from anyscale.models.service_model import ServiceConfig


def format_service_config(
    service_config: ServiceConfig,
) -> ApplyProductionServiceV2Model:
    if not service_config.ray_serve_config:
        raise RuntimeError(
            "ray_serve_config is expected for the Service configuration."
        )

    service_name = service_config.name

    return ApplyProductionServiceV2Model(
        name=service_name,
        description=service_config.description or "Service updated from CLI",
        project_id=service_config.project_id,
        version=service_config.version,
        canary_percent=service_config.canary_percent,
        ray_serve_config=service_config.ray_serve_config,
        ray_gcs_external_storage_config=service_config.ray_gcs_external_storage_config,
        build_id=service_config.build_id,
        compute_config_id=service_config.compute_config_id,
        rollout_strategy=service_config.rollout_strategy,
        config=service_config.config,
    )



from typing import Generator
from unittest.mock import Mock, patch

from click.testing import CliRunner
import pytest

from anyscale.commands.service_commands import rollout


@pytest.fixture()
def mock_service_controller() -> Generator[Mock, None, None]:
    mock_service_controller = Mock()

    with patch(
        "anyscale.commands.service_commands.ServiceController",
        new=mock_service_controller,
    ):
        yield mock_service_controller


def test_rollout_strategy(mock_service_controller):
    """Tests the logic for setting `rollout_strategy`.

    This can either be set by `--rollout-strategy` or the `-i / --in-place` alias (but not both).
    """
    runner = CliRunner()

    # --rollout-strategy provided directly.
    runner.invoke(rollout, args=["--rollout-strategy=IN_PLACE", "-f", "file"])
    mock_service_controller().rollout.assert_called_once_with(
        "file",
        name=None,
        version=None,
        canary_percent=None,
        rollout_strategy="IN_PLACE",
    )
    mock_service_controller().reset_mock()

    runner.invoke(rollout, args=["--rollout-strategy=ROLLOUT", "-f", "file"])
    mock_service_controller().rollout.assert_called_once_with(
        "file",
        name=None,
        version=None,
        canary_percent=None,
        rollout_strategy="ROLLOUT",
    )
    mock_service_controller().reset_mock()

    # -i / --in-place used.
    runner.invoke(rollout, args=["-i", "-f", "file"])
    mock_service_controller().rollout.assert_called_once_with(
        "file",
        name=None,
        version=None,
        canary_percent=None,
        rollout_strategy="IN_PLACE",
    )
    mock_service_controller().reset_mock()

    runner.invoke(rollout, args=["--in-place", "-f", "file"])
    mock_service_controller().rollout.assert_called_once_with(
        "file",
        name=None,
        version=None,
        canary_percent=None,
        rollout_strategy="IN_PLACE",
    )
    mock_service_controller().reset_mock()

    # Provided both -i / --in-place and --rollout-strategy should error.
    result = runner.invoke(
        rollout, args=["-i", "--rollout-strategy=ROLLOUT", "-f", "file"]
    )
    assert result.exception is not None

    result = runner.invoke(
        rollout, args=["--in-place", "--rollout-strategy=ROLLOUT", "-f", "file"]
    )
    assert result.exception is not None



from typing import Optional

from rich.console import Console

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models import ListMachinesResponse
from anyscale.cloud import get_cloud_id_and_name
from anyscale.controllers.base_controller import BaseController


class MachineController(BaseController):
    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log
        self.console = Console()

    def list_machines(
        self, cloud: Optional[str], cloud_id: Optional[str]
    ) -> ListMachinesResponse:
        if cloud is not None:
            cloud_id, _ = get_cloud_id_and_name(self.api_client, cloud_name=cloud)

        response: ListMachinesResponse = self.api_client.list_machines_api_v2_machines_get(
            cloud_id
        ).result
        return response



import enum
from typing import Any, Dict, Optional

from pydantic import Field, root_validator

from anyscale.models.job_model import (
    _validate_and_modify_runtime_env,
    BaseHAJobConfig,
)


class UserServiceAccessTypes(str, enum.Enum):
    private = "private"
    public = "public"


class ServiceConfig(BaseHAJobConfig):
    name: str = Field(
        ..., description="Name of service to be submitted.",
    )
    access: UserServiceAccessTypes = Field(
        UserServiceAccessTypes.public,
        description=(
            "Whether user service (eg: serve deployment) can be accessed by public "
            "internet traffic. If public, a user service endpoint can be queried from "
            "the public internet with the provided authentication token. "
            "If private, the user service endpoint can only be queried from within "
            "the same Anyscale cloud and will not require an authentication token."
        ),
    )

    ray_serve_config: Optional[Dict[str, Any]] = Field(
        None,
        description=(
            "The Ray Serve config to use for this Production service. It is supported only on v2 clouds. "
            "This config defines your Ray Serve application, and will be passed directly to Ray Serve. "
            "You can learn more about Ray Serve config files here: https://docs.ray.io/en/latest/serve/production-guide/config.html"
        ),
    )

    ray_gcs_external_storage_config: Optional[Dict[str, Any]] = Field(
        None,
        description=(
            "Address to connect to external storage at. "
            "Must be accessible from instances running in the provided cloud. "
            "This is only supported for v2 services."
        ),
    )

    # Version level fields for Services V2
    version: Optional[str] = Field(
        None,
        description="A version string that represents the version for this service. "
        "Will be populated with the hash of the config if not specified.",
    )
    canary_percent: Optional[int] = Field(
        None,
        description="A manual target weight for this service. "
        "If this field is not set, the service will automatically roll out. "
        "If set, this should be a number between 0 and 100 (inclusive). "
        "The newly created version will have weight `canary_percent` "
        "and the existing version will have `100 - canary_percent`.",
    )

    rollout_strategy: Optional[str] = Field(
        None,
        description="Strategy for rollout. "
        "The ROLLOUT strategy will deploy your Ray Serve configuration onto a newly started cluster, and then shift traffic over to the new cluster. "
        "You can manually control the speed of the rollout using the canary_weight configuration.\n"
        "The IN_PLACE strategy will use Ray Serve in place upgrade to update your existing cluster in place. "
        "When using this rollout strategy, you may only change the ray_serve_config field. "
        "You cannot partially shift traffic or rollback an in place upgrade. "
        "In place upgrades are faster and riskier than rollouts, and we recommend only using them for relatively safe changes (for example, increasing the number of replicas on a Ray Serve deployment).\n"
        "Default strategy is ROLLOUT.",
    )

    config: Optional[Dict[str, Any]] = Field(
        None, description="Target Service's configuration",
    )

    @root_validator
    def validates_config(cls, values) -> Dict[str, Any]:
        assert (
            values.get("runtime_env") is None
        ), "runtime_env should not be set for Services on v2 clouds."

        return values

    @root_validator
    def validate_runtime_env_v2(cls: Any, values: Any) -> Any:  # noqa: PLR0912
        ray_serve_config = values.get("ray_serve_config")
        if ray_serve_config is None:
            return values

        if "applications" in ray_serve_config:
            for ray_serve_app_config in ray_serve_config["applications"]:
                runtime_env = ray_serve_app_config.get("runtime_env")
                if runtime_env is not None:
                    _validate_and_modify_runtime_env(runtime_env)
        else:
            runtime_env = ray_serve_config.get("runtime_env")
            if runtime_env is not None:
                _validate_and_modify_runtime_env(runtime_env)

        return values



from anyscale.shared_anyscale_utils.utils.ray_semver import ray_semver_compare


def get_alias_for_base_image_name(base_image_name: str) -> str:
    """
    Modify a base image name to comply with Ray version and naming conventions.

    This function takes a base image name as input, extracts the Ray version, and checks if the version
    is greater than or equal to '2.7.0', or falls back to the original name.

    Args:
        base_image_name (str): The name of the base image to be checked.

    Returns:
        str: The modified base image name based on version and naming conventions.

    Example:
        To check a base image name, you can call this function like this:
        >>> get_alias_for_base_image_name("anyscale/ray:2.7.0-py38-cuda121"), it should be modified to "anyscale/ray:2.7.0oss-py38-cuda121"
    """
    # to extract "2.7.0", "2.7.0oss", "2.7.0optimized", "2.0.1rc" as ray_version
    ray_version = get_ray_version(base_image_name)

    if ray_version.endswith(("optimized", "oss")) or "rc" in ray_version:
        return base_image_name

    if ray_semver_compare(ray_version, "2.7.0") >= 0:  # ray_version >= "2.7.0"
        return base_image_name.replace(ray_version, f"{ray_version}optimized")
    return base_image_name


def get_ray_version(base_image: str) -> str:
    """Returns the Ray version in use based on the base image.

    Args:
        base_image: e.g. anyscale/ray-ml:1.9.0-cpu

    Returns:
        The ray version, e.g. 1.9.0.
    """
    # e.g. 1.9.0-cpu
    image_version = base_image.split(":")[-1]
    # e.g. 1.9.0
    ray_version = image_version.split("-")[0]
    return ray_version



import json
import os
import time


IDLE_TERMINATION_DIR = os.environ.get(
    "IDLE_TERMINATION_DIR", "/tmp/anyscale/idle_termination_reports"
)


def get_activity(file_path: str, reason: str):
    """
    A generic method to parse a JSON file and returns an object compatible with the Component Activity endpoint in Ray.

    The input file is of the following format
    {
        error: "some optional error string, if present the last_activity_timestamp is ignored",
        last_activity_timestamp: 1666226726.043797
    }

    If an error string exists, the is_active state is set to error with the error string is passed as a reason. Otherwise, the
    last_activity_timestamp & reason are passed as it, and is_active is set to INACTIVE.

    Parameters
    ----------
    file_path: str
    The file path of the activity file.

    reason: str
    The reason for the activity if the file exists.

    Returns
    -------
    dict[str, Any]
    The activity object
    """

    # If the file is not there return error
    if not os.path.exists(file_path):
        return {
            "is_active": "ERROR",
            "reason": "status file does not exist",
            "timestamp": time.time(),
        }
    try:
        with open(file_path) as f:
            result = json.loads(f.read())
            # check if an error exists
            error = result.get("error")
            if error:
                return {
                    "is_active": "ERROR",
                    "reason": error,
                    "timestamp": time.time(),
                }
            else:
                return {
                    "is_active": "INACTIVE",
                    "reason": reason,
                    "timestamp": time.time(),
                    "last_activity_at": result["last_activity_timestamp"],
                }
    except Exception:  # noqa: BLE001
        return {
            "is_active": "ERROR",
            "reason": "error parsing the status file",
            "timestamp": time.time(),
        }


def env_hook():
    """
    This is to be used by the RAY_CLUSTER_ACTIVITY_HOOK in order to report additional activities
    via the /api/component_activities API.
    """
    return {
        "workspace": get_activity(
            f"{IDLE_TERMINATION_DIR}/workspace.json", "workspace last snapshot"
        ),
        "web_terminal": get_activity(
            f"{IDLE_TERMINATION_DIR}/web_terminal.json", "web terminal command finished"
        ),
    }



# pylint:disable=private-import
import copy
import hashlib
import logging
import os
from pathlib import Path
import tempfile
from typing import Any, Callable, Dict, Optional, TYPE_CHECKING
from urllib.parse import urlparse

import click

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client import ComputeTemplate
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.client.openapi_client.models.cloud_providers import CloudProviders
from anyscale.client.openapi_client.models.cloud_with_cloud_resource import (
    CloudWithCloudResource,
)
from anyscale.client.openapi_client.models.cloud_with_cloud_resource_gcp import (
    CloudWithCloudResourceGCP,
)
from anyscale.client.openapi_client.models.user_info import UserInfo
from anyscale.shared_anyscale_utils.aws import bucket_name_from_maybe_bucket_arn
from anyscale.util import is_anyscale_workspace
from anyscale.utils.ray_utils import zip_directory  # type: ignore
from anyscale.utils.workload_types import Workload


if TYPE_CHECKING:
    pass

logger = logging.getLogger(__name__)


def _upload_file_to_google_cloud_storage(file: str, bucket: str, object_name: str):
    try:
        from google.cloud import storage

    except Exception:  # noqa: BLE001
        raise click.ClickException(
            "Could not upload file to Google Storage. Could not import the Google Storage Python API via `from google.cloud import storage`.  Please check your installation or try running `pip install --upgrade google-cloud-storage`."
        )
    try:
        storage_client = storage.Client()
        bucket_obj = storage_client.bucket(bucket)
        blob = bucket_obj.blob(object_name)
        blob.upload_from_filename(file)
    except Exception as e:  # noqa: BLE001
        raise click.ClickException(
            f"Failed to upload the working directory to Google Cloud Storage. Error {repr(e)}"
            "Please validate you have exported cloud credentials with the correct write permissions and the intended bucket exists in your Cloud Storage account. "
            "If you do not desire to upload your working directory, please set your working directory to a public remote URI or remove the runtime_environment from you service yaml."
        ) from e


def _upload_file_to_s3(file: str, bucket: str, object_key: str):
    try:
        import boto3
    except Exception:  # noqa: BLE001
        raise click.ClickException(
            "Could not upload file to S3: Could not import the Amazon S3 Python API via `import boto3`.  Please check your installation or try running `pip install boto3`."
        )
    try:
        s3_client = boto3.client("s3")
        s3_client.upload_file(file, bucket, object_key)
    except Exception as e:  # noqa: BLE001
        raise click.ClickException(
            f"Failed to upload the working directory to S3. Error {repr(e)}"
            "Please validate you have exported cloud credentials with the correct write permissions and the intended bucket exists in your S3 account. "
            "If you do not desire to upload your working directory, please set your working directory to a public remote URI or remove the runtime_environment from you service yaml."
        ) from e


def _get_remote_storage_object_name(upload_path, upload_filename):
    # Strip leading slash, otherwise bucket will create a new directory called "/".
    object_name = os.path.join(urlparse(upload_path).path, upload_filename).lstrip("/")
    return object_name


def _upload_file_to_remote_storage(
    source_file: str, upload_path: str, upload_filename: str
):
    parsed_upload_path = urlparse(upload_path)
    service = parsed_upload_path.scheme
    bucket = parsed_upload_path.netloc
    object_name = _get_remote_storage_object_name(upload_path, upload_filename)
    if service == "s3":
        _upload_file_to_s3(source_file, bucket, object_key=object_name)
    if service == "gs":
        _upload_file_to_google_cloud_storage(
            source_file, bucket, object_name=object_name
        )

    final_uploaded_filepath = os.path.join(upload_path, upload_filename)
    try:
        from smart_open import open

        open(final_uploaded_filepath)
    except Exception as e:  # noqa: BLE001
        raise click.ClickException(
            f"Could not open uploaded file, maybe something went wrong while uploading: {e}."
        )

    return final_uploaded_filepath


def is_dir_remote_uri(working_dir: str) -> bool:
    parsed = urlparse(working_dir)
    if parsed.scheme:
        return True
    return False


def upload_and_rewrite_working_dir(
    runtime_env_json: Dict[str, Any],
    upload_file_to_remote_storage_fn: Callable[
        [str, str, str], str
    ] = _upload_file_to_remote_storage,
) -> Dict[str, Any]:
    """Upload a local working_dir and rewrite the working_dir field with the destination remote URI.

    After uploading, deletes the "upload_path" field because it is no longer used and is not a valid
    OSS runtime env field.
    """
    if runtime_env_json.get("working_dir", None) is None:
        return runtime_env_json

    working_dir = runtime_env_json["working_dir"]
    if is_dir_remote_uri(working_dir):
        # The working dir is a remote URI already
        return runtime_env_json

    upload_path = runtime_env_json["upload_path"]
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_zip_file_path = os.path.join(
            temp_dir, "anyscale_generated_working_dir.zip"
        )
        zip_directory(
            working_dir,
            excludes=runtime_env_json.get("excludes", []),
            output_path=temp_zip_file_path,
            # Ray requires remote Zip URIs to consist of a single top-level directory when unzipped.
            include_parent_dir=True,
        )

        hash_val = hashlib.md5(Path(temp_zip_file_path).read_bytes()).hexdigest()
        uploaded_zip_file_name = f"_anyscale_pkg_{hash_val}.zip"
        final_uploaded_filepath = upload_file_to_remote_storage_fn(
            temp_zip_file_path, upload_path, uploaded_zip_file_name,
        )

    final_runtime_env = runtime_env_json.copy()
    final_runtime_env["working_dir"] = final_uploaded_filepath
    del final_runtime_env["upload_path"]
    return final_runtime_env


def override_runtime_env_config(
    runtime_env: Optional[Dict[str, Any]],
    anyscale_api_client: DefaultApi,
    api_client: DefaultApi,
    workload_type: Workload,
    compute_config_id: Optional[str],
    log: BlockLogger,
) -> Optional[Dict[str, Any]]:
    """ Override working_dir and upload_path for runtime environment
    1. If workspace environment, autopopulate working_dir
    2. If runtime env or working_dir is None or working_dir is a remote_uri, return existing runtime_env
    3. If working_dir is local directory and upload path is defined, upload and rewrite working_dir
    4. If working_dir is local directory and upload path is not defined, infer upload path, upload working_dir, and rewrite working_dir
    """

    existing_runtime_env = runtime_env

    existing_runtime_env = autopopulate_working_dir_for_workspace(
        runtime_env=existing_runtime_env, log=log
    )

    # If runtime is None, return empty dict
    # Elif working_dir is missing, return existing runtime_env
    if not existing_runtime_env:
        return {}
    elif not existing_runtime_env.get("working_dir"):
        return existing_runtime_env

    working_dir = existing_runtime_env.get("working_dir", "")
    upload_path = existing_runtime_env.get("upload_path")

    if not is_dir_remote_uri(working_dir):
        if upload_path is not None:
            # If upload_path is specified
            # we back up the current working dir to the specified path
            new_runtime_env = upload_and_rewrite_working_dir(existing_runtime_env)
        else:
            if is_anyscale_workspace() and "ANYSCALE_SESSION_ID" in os.environ:
                # If submitting job v2 from workspaces and no upload_path is specified,
                # we back up the current workspace content into S3
                cluster_id = os.environ["ANYSCALE_SESSION_ID"]

                decorated_cluster = api_client.get_decorated_cluster_api_v2_decorated_sessions_cluster_id_get(
                    cluster_id
                ).result
                cloud_id = decorated_cluster.cloud.id

                workspace_id = os.environ["ANYSCALE_EXPERIMENTAL_WORKSPACE_ID"]

                new_runtime_env = infer_upload_path_and_rewrite_working_dir(
                    api_client=api_client,
                    existing_runtime_env=existing_runtime_env,
                    workload_type=workload_type,
                    cloud_id=cloud_id,
                    log=log,
                    workspace_id=workspace_id,
                )
            else:
                compute_template: ComputeTemplate = anyscale_api_client.get_compute_template(
                    compute_config_id
                ).result
                cloud_id = compute_template.config.cloud_id
                new_runtime_env = infer_upload_path_and_rewrite_working_dir(
                    api_client=api_client,
                    existing_runtime_env=existing_runtime_env,
                    workload_type=workload_type,
                    cloud_id=cloud_id,
                    log=log,
                )

        return new_runtime_env
    else:
        return existing_runtime_env


def autopopulate_working_dir_for_workspace(
    runtime_env: Optional[Dict[str, Any]], log: BlockLogger,
) -> Optional[Dict[str, Any]]:
    """Set the default working_dir to local directory if deploying from workspace
    """
    if is_anyscale_workspace():
        if not runtime_env:
            runtime_env = {}

        if not runtime_env.get("working_dir"):
            runtime_env["working_dir"] = "."
            log.info(
                "Working_dir is not specified, setting your current local directory as the working_dir"
            )

    return runtime_env


def infer_upload_path_and_rewrite_working_dir(
    *,
    api_client: DefaultApi,
    existing_runtime_env: Dict[str, Any],
    cloud_id: str,
    workload_type: Workload,
    log: BlockLogger,
    workspace_id: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Push working_dir to remote bucket and rewrite the working_dir field with the destination uri

    If the upload_path is not specified by the user, we will get the bucket name based on the cloud.
    We then rewrite the working_dir to the remote uri path
    so that the launched service will read from remote bucket directly.

    For Workspaces:
        The remote path: [s3, gs]://{bucket_name}/{org_id}/{cloud_id}/workspace_snapshots/{workspace_id}/{workload_type}/{backup_zip}
    Otherwise:
        The remote path: [s3, gs]://{bucket_name}/{org_id}/{cloud_id}/{workload_type}/{backup_zip}
        workload_type=[jobs, scheduled_jobs, services]
    """

    cloud: CloudWithCloudResource = api_client.get_cloud_with_cloud_resource_api_v2_clouds_with_cloud_resource_router_cloud_id_get(
        cloud_id
    ).result
    org_id = _get_organization_id(api_client)

    if cloud.provider == CloudProviders.AWS:
        bucket_name = _get_cloud_s3_bucket_from_cloud(cloud)
        protocol = "s3"
    elif cloud.provider == CloudProviders.GCP:
        bucket_name = _get_cloud_gs_bucket_from_cloud(api_client, cloud)
        protocol = "gs"
    else:
        raise click.ClickException(
            f"Currently launching a service from workspaces in a {cloud.provider} cloud is not supported. "
            "Please contact Anyscale support for more info."
        )

    new_runtime_env = copy.deepcopy(existing_runtime_env)
    pre_upload_working_dir = new_runtime_env["working_dir"]

    if workspace_id:
        new_runtime_env[
            "upload_path"
        ] = f"{protocol}://{bucket_name}/{org_id}/{cloud_id}/workspace_snapshots/{workspace_id}/{workload_type}"
    else:
        new_runtime_env[
            "upload_path"
        ] = f"{protocol}://{bucket_name}/{org_id}/{cloud_id}/{workload_type}"

    new_runtime_env = upload_and_rewrite_working_dir(new_runtime_env)

    log.info(
        f"Uploaded working directory content from {Path(pre_upload_working_dir).absolute()} to {new_runtime_env['working_dir']}"
    )
    return new_runtime_env


def _get_organization_id(api_client: DefaultApi):
    user_info: UserInfo = (api_client.get_user_info_api_v2_userinfo_get().result)
    orgs = user_info.organizations
    return orgs[0].id


def _get_cloud_s3_bucket_from_cloud(cloud: CloudWithCloudResource) -> Optional[str]:
    """
    If the cloud has an associated aws s3 bucket, we return its name.

    Please note that this is only for v2 clouds where customers have their
    own S3 buckets.
    """
    assert cloud.provider == CloudProviders.AWS
    if cloud and cloud.cloud_resource and cloud.cloud_resource.aws_s3_id:
        return bucket_name_from_maybe_bucket_arn(cloud.cloud_resource.aws_s3_id)
    else:
        return None


def _get_cloud_gs_bucket_from_cloud(
    api_client: DefaultApi, cloud: CloudWithCloudResource
) -> Optional[str]:
    """
    If the cloud has an associated Google Storage bucket, we return its name.

    Please note that this is only for v2 clouds where customers have their
    own Google Storage.
    """
    assert cloud.provider == CloudProviders.GCP

    gcp_cloud: CloudWithCloudResourceGCP = api_client.get_cloud_with_cloud_resource_api_v2_clouds_with_cloud_resource_gcp_router_cloud_id_get(
        cloud.id
    ).result

    if (
        gcp_cloud
        and gcp_cloud.cloud_resource
        and gcp_cloud.cloud_resource.gcp_cloud_storage_bucket_id
    ):
        gs_bucket_name = gcp_cloud.cloud_resource.gcp_cloud_storage_bucket_id
        return gs_bucket_name
    else:
        return None
